# -*- coding:utf-8 -*-
from __future__ import absolute_import

import datetime
import sqlite3

import hashlib
import io
import numpy as np
import os
from sqlalchemy import Column, Integer, String, PrimaryKeyConstraint, Text, Boolean, DateTime, BINARY
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Session

from ..basic_tools.md5_tools import get_md5

Base = declarative_base()


def adapt_array(arr):
    """
    http://stackoverflow.com/a/31312102/190597 (SoulNibbler)
    """
    out = io.BytesIO()
    np.save(out, arr)
    out.seek(0)
    return sqlite3.Binary(out.read())  # todo only support sqlite3 ?


def convert_array(text):
    out = io.BytesIO(text)
    out.seek(0)
    return np.load(out)


def get_sqlite_engine(sqlite_file="tmp.dat"):
    """

    :rtype: Engine
    """
    return create_engine("sqlite:///{}".format(os.path.abspath(sqlite_file)),
                         echo=False).connect()


def get_sqlite_session(engine) -> Session:
    return sessionmaker(bind=engine)()


class RecordMixin(object):
    """保存account 信息"""

    # global var
    sqlite_file = "sqlite.dat"
    NULL_TIME = datetime.datetime(2000, 1, 1, 0, 0, 0)
    _engine = None

    @classmethod
    def create_table_by_code(cls):
        """
        创建表格
        :rtype: None
        """
        cls.__table__.create(cls.get_engine())

    @classmethod
    def get_engine(cls):
        """

        :rtype: Engine
        """
        if cls._engine is None:
            cls._engine = get_sqlite_engine(sqlite_file=cls.sqlite_file)

        return cls._engine

    @classmethod
    def get_new_session(cls) -> Session:
        return get_sqlite_session(engine=get_sqlite_engine(sqlite_file=cls.sqlite_file))

    @classmethod
    def destroy_new_session(cls, db_session: Session):
        if db_session is None:
            return
        db_session.bind.close()
        db_session.close()

    @classmethod
    def close_engine(cls):
        if cls._engine:
            cls._engine.close()
            cls._engine = None

    @classmethod
    def get_session(cls) -> Session:
        return get_sqlite_session(cls.get_engine())

    @classmethod
    def init_table_object(cls):
        if not cls.is_table_exists():
            cls.create_table_by_code()

    @classmethod
    def is_table_exists(cls):
        engine = cls.get_engine()
        return engine.dialect.has_table(engine.connect(), cls.__tablename__)


class FeatureRecord(Base, RecordMixin):
    """保存account 信息"""
    __tablename__ = 'feature'
    __table_args__ = (
        PrimaryKeyConstraint('id', name='pk_id'),
    )

    # pk
    id = Column(String(32), nullable=False, primary_key=True)
    image_id = Column(String(32))
    label = Column(Integer)
    feature = Column(BINARY)

    TEST_LABEL = -1

    @classmethod
    def get_info(cls, db_session: Session) -> (int, int):
        all_count = db_session.query(cls).count()
        assert isinstance(all_count, int)

        test_count = db_session.query(cls).filter(cls.label == cls.TEST_LABEL).count()
        assert isinstance(test_count, int)

        return all_count - test_count, test_count

    @classmethod
    def create_uid(cls, image_id: str, is_train: bool) -> str:
        raw = "{}:{}".format(image_id, is_train)
        return get_md5(raw.encode("utf-8"))

    @classmethod
    def add_feature(cls, db_session: Session, label: int, feature: np.ndarray, image_id: str):
        uid = cls.create_uid(image_id=image_id, is_train=bool(label != cls.TEST_LABEL))
        obj = db_session.query(cls).filter(cls.id == uid).first()
        if not obj:
            obj = cls()
            obj.id = uid

        obj.label = label
        obj.feature = adapt_array(feature)
        obj.image_id = image_id
        db_session.add(obj)
        db_session.commit()
        return obj

    @classmethod
    def list_feature(cls, db_session: Session, uid_list: list) -> list:
        result = db_session.query(cls).filter(cls.id.in_(uid_list))
        result_list = []
        for obj in result:
            result_list.append((obj.id, obj.image_id, obj.label, convert_array(obj.feature)))

        return result_list

    @classmethod
    def get_feature(cls, db_session: Session, uid: str):
        obj = db_session.query(cls).filter(cls.id == uid).first()
        if obj:
            return obj.id, obj.image_id, obj.label, convert_array(obj.feature)
        return None


class SpiderRecord(Base, RecordMixin):
    """保存account 信息"""
    __tablename__ = 'spider_record'
    __table_args__ = (
        PrimaryKeyConstraint('id', name='pk_id'),
    )

    # pk
    id = Column(Integer)
    uid = Column(String(32), nullable=False, unique=True)

    # record
    class_name = Column(String(128))
    url = Column(Text)
    added_time = Column(DateTime, nullable=False)
    account_count = Column(Integer)
    last_crawled_time = Column(DateTime, nullable=False)
    next_crawled_time = Column(DateTime, nullable=False)
    crawled_every_seconds = Column(Integer)

    # deleted_status
    deleted = Column(Boolean, default=False)
    fail_count = Column(Integer, default=0)
    MAX_FAIL_COUNT = 5

    # default crawl every_seconds
    CRAWLED_EVERY_SECONDS = 300

    @classmethod
    def get_next_task(cls, db_session):
        """

        :rtype: SpiderRecord|None
        :type db_session: Session
        """
        time_now = datetime.datetime.now()

        return db_session.query(cls) \
            .filter(cls.deleted == False, cls.next_crawled_time <= time_now) \
            .order_by(cls.next_crawled_time.asc()) \
            .first()

    @classmethod
    def recrawl_all_tasks(cls, db_session):
        """
        :type db_session: Session
        """
        time_now = datetime.datetime.now()

        db_session.query(cls) \
            .filter(cls.deleted == False, cls.next_crawled_time > time_now) \
            .update(dict(next_crawled_time=time_now))

        db_session.commit()

    @classmethod
    def from_target_url(cls, db_session, target_url):
        """

        :type db_session: Session
        :rtype: (SpiderRecord, is_created)
        :type target_url: str
        """
        class_name = "function"
        url = target_url.rstrip("/")

        hash_md5 = hashlib.md5("{}:{}".format(class_name, url).encode("utf-8"))
        uid = hash_md5.hexdigest()

        instance = db_session.query(cls).filter_by(uid=uid).first()
        if instance is None:
            now = datetime.datetime.now()
            instance = cls()
            instance.uid = uid
            instance.class_name = class_name
            instance.url = url
            instance.added_time = now
            instance.account_count = 0
            instance.last_crawled_time = cls.NULL_TIME
            instance.next_crawled_time = now
            instance.crawled_every_seconds = cls.CRAWLED_EVERY_SECONDS
            instance.deleted = False
            instance.fail_count = 0
            return instance, True
        else:
            return instance, False

    def commit_crawled_record(self, db_session, is_success=True):
        """

        :type is_success: bool
        :type db_session: Session
        """
        self.last_crawled_time = datetime.datetime.now()
        self.next_crawled_time = self.last_crawled_time \
                                 + datetime.timedelta(seconds=self.crawled_every_seconds)
        if is_success:
            self.fail_count = 0
        else:
            self.fail_count += 1
            if self.fail_count > self.MAX_FAIL_COUNT:
                self.deleted = True

        db_session.add(self)
        db_session.commit()

    @classmethod
    def delete_all_tasks(cls, db_session):
        """
        将所有使用中的账号设置为未使用
        :rtype: bool
        """
        db_session.query(cls) \
            .filter_by(deleted=False) \
            .update(dict(deleted=True))
        db_session.commit()

    @classmethod
    def reactive_task(cls, instance):
        now = datetime.datetime.now()
        instance.next_crawled_time = now
        instance.deleted = False
        instance.fail_count = 0


__all__ = ("Base", "get_sqlite_engine", "get_sqlite_session", "RecordMixin",
           "FeatureRecord", "adapt_array", "convert_array")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import os
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from typing import NamedTuple


class SourceEmailInfo(NamedTuple):
    """ email info """
    email_address: str
    email_password: str
    smtp_host: str
    smtp_port: int


def send_email(source_email_info: SourceEmailInfo, to_email_list: [str], title: str, content: str, file_name_list: [str] = None, ):
    msg = MIMEMultipart()

    # 基本信息
    msg['From'] = source_email_info.email_address
    msg['To'] = ",".join(to_email_list)
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = title

    # 邮件正文
    msg.attach(MIMEText(content, 'plain', 'utf-8'))

    # 添加附件
    if file_name_list:
        # todo: gbk, only for excel file
        for file_name in file_name_list:
            excel_file_obj = MIMEApplication(open(file_name, 'rb').read())
            excel_file_obj.add_header('Content-Disposition', 'attachment', filename=('gbk', '', os.path.basename(file_name)))
            msg.attach(excel_file_obj)

    # 发邮件
    smtp_obj = smtplib.SMTP_SSL(source_email_info.smtp_host, source_email_info.smtp_port)
    smtp_obj.login(source_email_info.email_address, source_email_info.email_password)
    smtp_obj.sendmail(source_email_info.email_address, to_email_list, msg.as_string())
    smtp_obj.quit()




# coding:utf-8
__author__ = 'rk.feng'

import asyncio
import gzip
import logging  # noqa
import os  # noqa
import random
import time
import typing
import unittest
from asyncio import futures
from collections import OrderedDict
from pprint import pprint
from urllib.parse import quote

import tornado.websocket
from tornado.websocket import WebSocketClientConnection, WebSocketClosedError

from pyxtools.basic_tools.md5_tools import create_guid, get_md5

REQUIRE_TEXT = """
pip install tornado>=5.1
"""


def build_device_id(route_key: str, device_hash: str = None) -> str:
    device_hash = device_hash or create_guid()
    return get_md5(route_key + "." + route_key[:-1])[:10] + get_md5(device_hash)[:5]


def check_device_id(route_key: str, device_id: str) -> bool:
    if len(route_key) < 3:
        return False
    if len(device_id) != 15:
        return False
    if get_md5(route_key + "." + route_key[:-1])[:10] != device_id[:10]:
        return False
    return True


class MessageBytesUtils:
    ws_id_byte_size = 2  # 16bit
    msg_no_size = 4  # 32bit
    frame_size = ws_id_byte_size + msg_no_size

    @classmethod
    def build_frame(cls, ws_id: int, msg_no: int) -> bytes:
        """ """
        return ws_id.to_bytes(MessageBytesUtils.ws_id_byte_size, byteorder='little', signed=False) + \
               msg_no.to_bytes(MessageBytesUtils.msg_no_size, byteorder='little', signed=False)

    @classmethod
    def parse_frame(cls, frame: bytes) -> typing.Tuple[int, int]:
        """ """
        return int.from_bytes(frame[:MessageBytesUtils.ws_id_byte_size], byteorder="little", signed=False), \
               int.from_bytes(frame[MessageBytesUtils.ws_id_byte_size:], byteorder="little", signed=False)

    @classmethod
    def build_message(cls, frame: bytes, msg: bytes) -> bytes:
        return frame[:MessageBytesUtils.frame_size] + msg

    @classmethod
    def parse_message(cls, message: bytes) -> typing.Tuple[bytes, bytes]:
        return message[:MessageBytesUtils.frame_size], message[MessageBytesUtils.frame_size:]

    @classmethod
    def compress_gzip(cls, msg_str: str) -> bytes:
        """ """
        return gzip.compress(msg_str.encode("utf-8"))

    @classmethod
    def decompress_gzip(cls, compressed_msg: bytes) -> str:
        """ """
        return gzip.decompress(compressed_msg).decode("utf-8")


class _RecvObj:
    __slots__ = ("fut", "waited")

    def __init__(self, fut: asyncio.Future, waited: float = 0, ):
        self.waited = waited  # waited的时间戳, <= 0 表示 还没被占用
        self.fut = fut

    @property
    def weight(self):
        had_data = bool(self.fut.cancelled() or self.fut.done())

        return 1 if had_data else 0, self.waited if self.waited > 0 else 0


class WsP2PClient:
    def __init__(self, route_key: str, ws_url: str, logger: logging.Logger, device_hash: str = None):
        self.logger = logger

        self._ws_server_conn: WebSocketClientConnection = None
        self.device_id: str = build_device_id(route_key, device_hash)
        self.route_key: str = route_key
        self._send_cache: typing.OrderedDict[bytes, asyncio.Future] = OrderedDict()
        self._recv_cache: typing.List[_RecvObj] = []
        self.ws_url = ws_url + f"?device_id={quote(self.device_id)}&route_key={quote(self.route_key)}"
        self.ws_id: int = None
        self.msg_no: int = 0

        self._lock = asyncio.Lock()

    def _create_frame_id(self, ) -> bytes:
        try:
            return MessageBytesUtils.build_frame(ws_id=self.ws_id, msg_no=self.msg_no)
        finally:
            self.msg_no += 1

    async def _open(self):
        """ """
        if self._ws_server_conn is not None:
            return

        async with self._lock:
            if self._ws_server_conn is not None:
                return

            self.logger.info(f"route_key : {self.route_key}, device_id {self.device_id}")
            self._ws_server_conn = await tornado.websocket.websocket_connect(
                self.ws_url,
                connect_timeout=60,
                ping_interval=10,
                on_message_callback=self._on_message_callback,
            )

            # check self.ws_id
            for _ in range(10):
                if self.ws_id is None:
                    await asyncio.sleep(1)
            if self.ws_id is None:
                raise ValueError("self.ws_id cannot be null!")

    def _on_message_callback(self, message: bytes):
        async def on_message_async(msg: bytes):
            # self.logger.info(f"on_message {message}")
            frame_id, msg = MessageBytesUtils.parse_message(message=msg)
            if self.ws_id is None:
                ws_id, msg_no = MessageBytesUtils.parse_frame(frame=frame_id)
                if ws_id == 0 and msg_no == 0:
                    self.ws_id = int.from_bytes(msg, byteorder="little", signed=False)
                    self.logger.info(f"[_on_message_callback][ws_id {self.ws_id}]ws_id set!")
                return

            if frame_id in self._send_cache:
                # response
                if not self._send_cache[frame_id].cancelled() and not self._send_cache[frame_id].done():
                    self._send_cache[frame_id].set_result(msg)
            else:
                # request
                fut: asyncio.Future = None
                recv_obj_list = sorted(
                    [recv_obj for recv_obj in self._recv_cache if not recv_obj.fut.done() or not recv_obj.fut.cancelled()],
                    key=lambda x: x.weight, reverse=True
                )
                if recv_obj_list:
                    recv_obj = recv_obj_list[0]
                    fut = recv_obj.fut
                else:
                    recv_obj = _RecvObj(fut=asyncio.Future())
                    fut = recv_obj.fut
                    self._recv_cache.append(recv_obj)

                if not fut.done() and not fut.cancelled():
                    fut.set_result((frame_id, msg))

        asyncio.ensure_future(on_message_async(message))

    async def send(self, msg: bytes, timeout: int = 0) -> bytes:
        await self._open()
        assert self.ws_id is not None
        frame_id = self._create_frame_id()
        self._send_cache[frame_id] = asyncio.Future()
        if self._ws_server_conn is None:
            raise ValueError("websocket closed!")

        try:
            bin_message = MessageBytesUtils.build_message(
                frame=frame_id,
                msg=msg,
            )
            await self._ws_server_conn.write_message(message=bin_message, binary=True)
        except WebSocketClosedError:
            self._ws_server_conn = None
            raise ValueError("websocket closed!")

        if timeout and timeout > 0:
            try:
                res = await asyncio.wait_for(self._send_cache[frame_id], timeout=timeout)
            except futures.TimeoutError as e:
                if not self._send_cache[frame_id].cancelled() or not self._send_cache[frame_id].done():
                    self._send_cache[frame_id].set_exception(e)
                raise
        else:
            res = await self._send_cache[frame_id]
            del self._send_cache[frame_id]

        return res

    async def receive(self, handle_msg: typing.Callable[[bytes, bytes], typing.Awaitable[bytes]]):
        await self._open()
        assert self.ws_id is not None

        # get req
        fut: asyncio.Future = None
        recv_obj_list = sorted(
            [recv_obj for recv_obj in self._recv_cache if recv_obj.waited <= 0], key=lambda x: x.weight, reverse=True
        )
        if recv_obj_list:
            recv_obj = recv_obj_list[0]
            recv_obj.waited = time.time()
            fut = recv_obj.fut
        else:
            recv_obj = _RecvObj(fut=asyncio.Future(), waited=time.time())
            self._recv_cache.append(recv_obj)
            fut = recv_obj.fut

        frame_id, msg = await fut
        _remove_index = None
        for recv_i, recv_obj in enumerate(self._recv_cache):
            if id(recv_obj.fut) == id(fut):
                _remove_index = recv_i
                break
        if _remove_index is not None:
            self._recv_cache.pop(_remove_index)

        # handle req
        if handle_msg:
            resp = await handle_msg(frame_id, msg)
        else:
            resp = b""

        # send resp
        if self._ws_server_conn is None:
            raise ValueError("websocket closed!")

        try:
            bin_message = MessageBytesUtils.build_message(
                frame=frame_id,
                msg=resp,
            )
            await self._ws_server_conn.write_message(message=bin_message, binary=True)
        except WebSocketClosedError:
            self._ws_server_conn = None
            raise ValueError("websocket closed!")


class T(unittest.TestCase):
    def setUp(self) -> None:
        logger = logging.getLogger()

        logger.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        self.logger = logger
        self.route_key = 'aaaxbbb2'
        self.ws_url = "wss://www.example.com/ws/p2p"

    def testServer(self):
        async def consume(frame_id: bytes, msg: bytes) -> bytes:
            await asyncio.sleep(random.randint(1, 3))
            msg_str = MessageBytesUtils.decompress_gzip(compressed_msg=msg)
            pprint(msg_str)
            return MessageBytesUtils.compress_gzip(msg_str=msg_str + " done!")

        async def s():
            x = WsP2PClient(route_key=self.route_key, ws_url=self.ws_url, logger=self.logger)
            while True:
                await x.receive(handle_msg=consume)

        asyncio.get_event_loop().run_until_complete(s())

    def testClient(self):
        async def send(c, msg: str, timeout=5) -> str:
            try:
                res_bytes = await c.send(MessageBytesUtils.compress_gzip(msg_str=msg), timeout=timeout)
                return MessageBytesUtils.decompress_gzip(res_bytes)
            except Exception as e:
                pass

            return ""

        async def c():
            x = WsP2PClient(route_key=self.route_key, ws_url=self.ws_url, logger=self.logger)
            i = 0
            while True:
                v_list = await asyncio.gather(*[
                    send(x, f"msg.{i + 1}"),
                    send(x, f"msg.{i + 2}"),
                ])
                for v in v_list:
                    pprint(v)
                i += 2

        asyncio.get_event_loop().run_until_complete(c())



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import pickle
import sqlite3
import time

import os

logger = logging.getLogger(__name__)


class AbstractCache(object):
    def __init__(self, cache_file: str):
        self.cache_file = cache_file

    def get(self, key: str) -> object:
        return self._get(key)

    def set(self, key: str, value: object, ttl: int = 1e10):
        return self._set(key, value, ttl)

    def unsafe_set(self, key: str, value: object, ttl: int = 1e10):
        return self._unsafe_set(key, value, ttl)

    def _set(self, key: str, value: object, ttl: int):
        raise NotImplemented

    def _unsafe_set(self, key: str, value: object, ttl: int):
        raise NotImplemented

    def _get(self, key: str) -> object:
        raise NotImplemented

    @staticmethod
    def _get_exp_time(ttl: int) -> int:
        return int(time.time() + ttl)

    @staticmethod
    def _ttl_is_timeout(exp_time: float) -> bool:
        return time.time() > exp_time


class SqliteCache(AbstractCache):
    def __init__(self, cache_file: str):
        super(SqliteCache, self).__init__(cache_file)

    def _get(self, key: str) -> object:
        conn = self._get_connect()
        try:
            value, exp_time = self._get_uid_value(conn, key)
            if exp_time is not None and not self._ttl_is_timeout(exp_time):
                return value

        except Exception as e:
            print(e)
        finally:
            conn.close()

        return None

    @staticmethod
    def _get_uid_value(conn, key):
        try:
            sql = "SELECT VALUE, EXPTIME from CACHE where UID = ?"
            cursor = conn.cursor().execute(sql, (key,))
            for row in cursor:
                return pickle.loads(row[0]), row[1]
        except Exception as e:
            logger.error(e)

        return None, None

    def _set(self, key: str, value: object, ttl: int):
        conn = self._get_connect()
        try:
            old_value, exp_time = self._get_uid_value(conn, key)
            if old_value is None and exp_time is None:
                sql = "INSERT INTO CACHE(ID, VALUE, EXPTIME, UID) VALUES (NULL, ?, ?, ?)"
            else:
                sql = "UPDATE CACHE SET VALUE = ?, EXPTIME = ? where UID = ?"

            conn.cursor().execute(sql, (pickle.dumps(value), self._get_exp_time(ttl), key))
            conn.commit()
        except Exception as e:
            logger.error(e)
            return None
        finally:
            conn.close()

    def _unsafe_set(self, key: str, value: object, ttl: int):
        logger.warning("unsafe_set == set in {}".format(self.__class__.__name__))
        return self._set(key, value, ttl)

    def _get_connect(self):
        if not os.path.exists(self.cache_file):
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute('''CREATE TABLE CACHE
                   (ID INTEGER PRIMARY KEY AUTOINCREMENT,
                   UID CHAR(32) UNIQUE NOT NULL,
                   EXPTIME INT NOT NULL ,
                   VALUE TEXT);''')
            conn.commit()
            conn.close()

        return sqlite3.connect(self.cache_file)


class FileCache(AbstractCache):
    def __init__(self, pickle_file: str):
        super(FileCache, self).__init__(pickle_file)
        self.cache = self._read()

    def _read(self) -> dict:
        if not os.path.exists(self.cache_file):
            return {}

        with open(self.cache_file, "rb") as f:
            return pickle.load(f)

    def _write(self, cache):
        with open(self.cache_file, "wb") as f:
            pickle.dump(cache, f)

    def _get(self, key: str) -> object:
        val_exp = self.cache.get(key)
        if val_exp is not None:
            value, exp_time = val_exp
            if not self._ttl_is_timeout(exp_time):
                return value
            else:
                self._set(key, None, 0)

        return None

    def _set(self, key: str, value: object, ttl: int):
        self._unsafe_set(key, value, ttl)
        self._write(self.cache)

    def _unsafe_set(self, key: str, value: object, ttl: int):
        if value is None and key in self.cache:
            self.cache.pop(key)
        else:
            self.cache[key] = (value, self._get_exp_time(ttl))


__all__ = ("SqliteCache", "FileCache", "AbstractCache")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import asyncio
from concurrent.futures import ThreadPoolExecutor

import aiohttp
import requests


def async_get(url_list, max_cpu=None, max_request_per_cpu=None, callback_func=None):
    async def get_response_content(url, session):
        async with session.get(url) as response:
            return await response.read()

    async def bound_fetch(sem, url, session, callback):
        # Getter function with semaphore.
        async with sem:
            content = await get_response_content(url, session)
            callback(url, content)
            return

    async def run(_url_list, sem):
        tasks = []

        # Create client session that will ensure we dont open new connection
        # per each request.
        async with aiohttp.ClientSession() as session:
            for url in _url_list:
                task = asyncio.ensure_future(bound_fetch(sem, url, session, callback_func))
                tasks.append(task)

            await asyncio.gather(*tasks, return_exceptions=True)

    loop = asyncio.get_event_loop()
    future = asyncio.ensure_future(
        run(
            url_list,
            sem=asyncio.Semaphore(max_request_per_cpu if max_request_per_cpu > len(url_list) else len(url_list))
        ))
    loop.run_until_complete(future)


def async_request_get(url_list, time_out=None, proxy=None, sema_count=None, response_callback=None):
    """

    :rtype: list of str|Exception|bytes
    :type response_callback: None|callable
    :type sema_count: None|int
    :type proxy: Nont|str
    :type time_out: None|int
    :type url_list: list of str
    """

    def sorted_response(response_list):
        return [data for (index, data) in sorted(response_list, key=lambda index_data: index_data[0])]

    if sema_count and sema_count > len(url_list):
        sema_count = len(url_list)

    # aiohttp only support http
    if proxy is None or proxy.startswith("http://"):
        async def get_async(url, index):
            if time_out is None:
                try:
                    async with aiohttp.request('GET', url, proxy=proxy, ) as r:
                        data = await r.read()
                        if response_callback:
                            data = response_callback(data)
                    return index, data
                except Exception as e:
                    return index, e
            else:
                timeout = aiohttp.ClientTimeout(total=time_out)
                try:
                    async with aiohttp.ClientSession(timeout=timeout) as session:
                        async with session.get(url, timeout=timeout, proxy=proxy) as r:
                            data = await r.read()
                            if response_callback:
                                data = response_callback(data)
                    return index, data
                except Exception as e:
                    return index, e

        if sema_count is None:
            # not use sema
            event_loop = asyncio.get_event_loop()
            tasks = [get_async(url, index) for index, url in enumerate(url_list)]
            results = event_loop.run_until_complete(asyncio.gather(*tasks))
            return sorted_response(results)

        # use sema
        sema = asyncio.Semaphore(sema_count)
        result_list = []

        async def dealwith_result(url, index):
            with (await sema):
                index, data = await get_async(url, index)
                result_list.append((index, data))

        loop = asyncio.get_event_loop()
        f = asyncio.wait([dealwith_result(url, index) for index, url in enumerate(url_list)])
        loop.run_until_complete(f)

        return sorted_response(result_list)

    else:
        # threadpool + requests
        def get_block(url_index):
            url, index = url_index
            try:
                data = requests.get(
                    url,
                    timeout=time_out,
                    proxies={'http': proxy, 'https': proxy} if proxy else None
                ).content
                if response_callback:
                    data = response_callback(data)
                return index, data
            except Exception as e:
                return index, e

        with ThreadPoolExecutor(max_workers=sema_count) as executor:
            result_list = [(index, data) for index, data in
                           executor.map(get_block, ((url, index) for index, url in enumerate(url_list)))]

        return sorted_response(result_list)


__all__ = ("async_request_get", "async_get")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import math
import time
import unittest

import numpy as np

from pyxtools.basic_tools.log import global_init_logger
from pyxtools.basic_tools.numpy_tools import NormType, calc_distance_pairs, _calc_distance_pairs_by_scipy

global_init_logger()


class TestNumpy(unittest.TestCase):
    def setUp(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def testL2Norm(self):
        input_data = np.asarray([[1.0, 2, 3, 1], [4.0, 5, 6, 4], [7.0, 8, 9, 7]])
        self.assertTrue(input_data.shape == (3, 4))

        # l2 [m, d], have d norm
        # norm np
        norm_list = [np.sqrt(np.dot(input_data[:, i], input_data[:, i].T)) for i in range(input_data.shape[-1])]
        norm_np = np.zeros(shape=input_data.shape)
        for i in range(input_data.shape[-1]):
            norm_np[:, i] = input_data[:, i] / norm_list[i]

        self.logger.info("norm np is {}".format(norm_np))
        self.assertTrue((abs(NormType.l2.normalize(input_data) - norm_np) < 1e-4).all())

        # l2: [m, d], have 1 norm
        # norm np
        total_norm = np.sqrt(np.sum(np.power(input_data, 2)))
        self.assertEqual(total_norm.shape, ())
        norm_np = input_data / total_norm
        self.logger.info("norm np is {}".format(norm_np))
        self.assertTrue((abs(norm_np - NormType.all.normalize(input_data)) < 1e-4).all())

    def testNormTypeReturnNorm(self):
        input_data = np.asarray([[1.0, 2, 3, 1], [4.0, 5, 6, 4], [7.0, 8, 9, 7]])
        self.assertTrue(input_data.shape == (3, 4))

        feature, norm = NormType.all.normalize_and_return_norm(input_data)
        self.assertTrue(isinstance(norm, float))

        feature, norm = NormType.l2.normalize_and_return_norm(input_data)
        self.assertTrue(isinstance(norm, np.ndarray))

        feature, norm = NormType.none.normalize_and_return_norm(input_data)
        self.assertTrue(norm is None)

    def testDistance(self):
        m, n, k = 5, 512, 25
        vec1 = np.random.random((m, n))
        for i in range(n):
            vec1[0][i] = 1
            vec1[1][i] = 0

        vec2 = np.concatenate((vec1, np.random.random((k, n))), axis=0)

        self.assertTrue(vec1.shape == (m, n))
        self.assertTrue(vec2.shape == (m + k, n))
        self.assertTrue((vec1 == vec2[:m]).all())
        self.assertTrue(np.sum(vec1[0]) == n)
        self.assertTrue(np.sum(vec1[1]) == 0)

        # dumpy
        distance_dumpy = np.zeros(shape=(m, m + k), dtype=np.float32)
        for i in range(m):
            for j in range(m + k):
                distance_dumpy[i][j] = np.sqrt(np.sum(np.square(vec1[i] - vec2[j])))
        self.assertTrue(distance_dumpy.shape == (m, m + k))
        self.assertTrue(distance_dumpy[0][0] == 0)
        self.assertTrue(distance_dumpy[1][1] == 0)
        self.assertTrue(abs(distance_dumpy[0][1] - math.sqrt(n)) < 1e-4)
        self.assertTrue(abs(distance_dumpy[1][0] - math.sqrt(n)) < 1e-4)

        # scipy: np.zeros((m, m + k))
        _time_start = time.time()
        for _ in range(10):
            distance_scipy = _calc_distance_pairs_by_scipy(vec1, vec2)
        t1 = time.time() - _time_start
        # self.logger.info("distance is {}".format(distance_scipy))
        # self.logger.info("distance_dumpy is {}".format(distance_dumpy))

        self.assertTrue(distance_scipy.shape == (m, m + k))
        self.assertTrue((abs(distance_dumpy - distance_scipy) < 1e-5).all())

        # numpy
        _time_start = time.time()
        for _ in range(10):
            distance_np = calc_distance_pairs(vec1, vec2)
        t2 = time.time() - _time_start

        self.assertTrue(distance_np.shape == (m, m + k))
        self.logger.info("scipy time is {}, numpy time is {}".format(t1, t2))
        self.assertGreater(t1, t2)
        self.assertTrue(self._compare_numpy(distance_dumpy, distance_np, 1e-4))
        self.assertTrue(self._compare_numpy(distance_scipy, distance_np, 1e-4))

    def testLen(self):
        a = np.random.random((5, 4))
        self.assertTrue(a.shape == (5, 4))
        self.assertTrue(len(a) == a.shape[0])

        a = np.random.random((5, 1))
        self.assertTrue(a.shape == (5, 1))
        self.assertTrue(len(a) == a.shape[0])

        a = np.random.random((5,))
        self.assertTrue(a.shape == (5,))
        self.assertTrue(len(a) == a.shape[0])

    def _compare_numpy(self, vec1, vec2, distance: float) -> bool:
        self.assertEqual(vec1.shape, vec2.shape)
        self.assertEqual(len(vec1.shape), 2)

        try:
            success = (abs(vec1 - vec2) <= distance).all()
            if success:
                return True
        except Exception as e:
            self.logger.error(e)

        self.logger.info("vec1 is {}".format(vec1))
        self.logger.info("vec2 is {}".format(vec2))

        fail_info = []
        shape = vec1.shape
        for i in range(shape[0]):
            for j in range(shape[1]):
                if abs(vec1[i][j] - vec2[i][j]) > distance:
                    fail_info.append("[{}, {}] {} vs {}".format(i, j, vec1[i][j], vec2[i][j]))

        if fail_info:
            self.logger.error("numpy is not equal:\n" + "\n".join(fail_info))
        return False



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import numpy as np

try:
    from pyxtools.faiss_tools import faiss
except ImportError:
    from pyxtools.pyxtools.faiss_tools import faiss


def test_faiss():
    d = 64  # dimension
    nb = 100000  # database size
    nq = 10000  # nb of queries
    np.random.seed(1234)  # make reproducible
    xb = np.random.random((nb, d)).astype('float32')
    xb[:, 0] += np.arange(nb) / 1000.
    xq = np.random.random((nq, d)).astype('float32')
    xq[:, 0] += np.arange(nq) / 1000.

    print("xq shape is {}, xb shape is {}".format(xq.shape, xb.shape))

    index = faiss.IndexFlatL2(d)  # build the index
    print(index.is_trained)
    index.add(xb)  # add vectors to the index
    print(index.ntotal)

    k = 4  # we want to see 4 nearest neighbors
    distance_list, indices = index.search(xb[:5], k)  # sanity check
    print(indices)
    print(distance_list)
    distance_list, indices = index.search(xq, k)  # actual search
    print(indices[:5])  # neighbors of the 5 first queries
    print(indices[-5:])  # neighbors of the 5 last queries


def faiss_learning():
    dimension = 2

    vec_data = np.random.random((9, dimension)).astype(np.float32)
    index = 0
    for x in [-1, 0, 1]:
        for y in [1, 0, -1]:
            vec_data[index, :] = np.array([x, y])
            index += 1

    vec_query = np.array([[0.1, 0], [0.0, 0.0], [1.0, 1.0], [2.0, 2.0], [3.0, 3.0]], dtype=np.float32)
    vec_query.reshape((5, dimension))

    print("vec_data shape is {}, vec_query shape is {}".format(vec_data.shape, vec_query.shape))

    index = faiss.IndexFlatL2(dimension)  # build the index
    print(index.is_trained)
    index.add(vec_data)  # add vectors to the index
    print(index.ntotal)

    k = 4  # we want to see 4 nearest neighbors
    distance_list, indices = index.search(vec_data[:5], k)  # sanity check
    print("distance_list is {}, indices is {}".format(distance_list, indices))

    distance_list, indices = index.search(vec_query, k)  # sanity check
    print("neighbors of the 5 first queries is {}, neighbors of the 5 last queries is {}"
          .format(indices[:5], indices[-5:]))


def faiss_kmeans():
    d = 3  # dimension
    nb = 100000  # database size
    np.random.seed(1234)  # make reproducible
    x = np.random.random((nb, d)).astype('float32')
    num_centroids = 1024
    niter = 20
    verbose = True
    d = x.shape[1]
    kmeans = faiss.Kmeans(d, num_centroids, niter, verbose)
    kmeans.train(x)

    index = faiss.IndexFlatL2(d)
    index.add(x)
    distance_list, indices = index.search(kmeans.centroids, 15)
    print("distance is {}".format(distance_list[:10]))
    print("distance is {}".format(distance_list[-10:]))


def faiss_pca():
    mt = np.random.rand(1000, 40).astype('float32')
    mat = faiss.PCAMatrix(40, 10)
    mat.train(mt)
    assert mat.is_trained
    tr = mat.apply_py(mt)
    print((tr ** 2).sum(0))


if __name__ == '__main__':
    faiss_pca()



# Copyright (c) 2015-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD+Patents license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np

try:
    from pyxtools.faiss_tools import faiss
except ImportError:
    from pyxtools.pyxtools.faiss_tools import faiss


def test_1():
    d = 64  # dimension
    nb = 100000  # database size
    nq = 10000  # nb of queries
    np.random.seed(1234)  # make reproducible
    xb = np.random.random((nb, d)).astype('float32')
    xb[:, 0] += np.arange(nb) / 1000.
    xq = np.random.random((nq, d)).astype('float32')
    xq[:, 0] += np.arange(nq) / 1000.

    nlist = 100
    m = 8
    k = 4
    quantizer = faiss.IndexFlatL2(d)  # this remains the same
    index = faiss.IndexIVFPQ(quantizer, d, nlist, m, 8)
    # 8 specifies that each sub-vector is encoded as 8 bits
    index.train(xb)
    index.add(xb)
    distance_list, indices = index.search(xb[:5], k)  # sanity check
    print(indices)
    print(distance_list)

    index.nprobe = 10  # make comparable with experiment above
    distance_list, indices = index.search(xq, k)  # search
    print(indices[:5])
    print(distance_list[:5])


def test_2():
    d = 64  # dimension
    nb = 100000  # database size
    nq = 10000  # nb of queries
    np.random.seed(1234)  # make reproducible
    xb = np.random.random((nb, d)).astype('float32')
    xb[:, 0] += np.arange(nb) / 1000.
    xq = np.random.random((nq, d)).astype('float32')
    xq[:, 0] += np.arange(nq) / 1000.

    nlist = 100
    m = 8
    k = 4
    quantizer = faiss.IndexFlatL2(d)  # this remains the same
    index = faiss.IndexIVFPQ(quantizer, d, nlist, m, 8)
    # 8 specifies that each sub-vector is encoded as 8 bits
    index.train(xb)
    index.add(xb)
    distance_list, indices = index.search(xb[:5], k)  # sanity check
    print(indices)
    print(distance_list)

    index.nprobe = 10  # make comparable with experiment above
    distance_list, indices = index.search(xq, k)  # search
    print(indices[:5])
    print(distance_list[:5])


if __name__ == '__main__':
    test_1()
    test_2()



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import os
import unittest

from pyxtools.basic_tools import get_pretty_float
from pyxtools.basic_tools.cache_tools import FileCache
from pyxtools.basic_tools.file_tools import get_base_name_of_file
from pyxtools.basic_tools.iterator_utils import random_choice, iter_list_with_size


class TestBasicTools(unittest.TestCase):
    def test_iter_list_with_size(self):
        src_list = [i for i in range(100)]
        raw_src_list = list(src_list)

        dst_list = []
        for part_list in iter_list_with_size(src_list, 20):
            dst_list.extend(part_list)

        self.assertEqual(len(raw_src_list), len(dst_list))
        self.assertEqual(sum(raw_src_list), sum(dst_list))
        self.assertEqual(len(src_list), 0)

    def test_cache(self):
        cache_file = "./cache.pkl"
        cache = FileCache(pickle_file=cache_file)
        test_dict = {"a": "b", "A": 0}

        # simple set
        for key, value in test_dict.items():
            cache.set(key, value)

        # test get
        self.assertEqual(cache.get("a"), test_dict.get("a"))
        self.assertEqual(cache.get("A"), test_dict.get("A"))
        self.assertEqual(cache.get("c"), test_dict.get("c"))

        # unsafe set
        cache.unsafe_set("X", "x")
        self.assertEqual(cache.get("X"), "x")

        # unsafe
        another = FileCache(pickle_file=cache_file)
        self.assertEqual(another.get("a"), test_dict.get("a"))
        self.assertEqual(another.get("A"), test_dict.get("A"))
        self.assertEqual(another.get("c"), test_dict.get("c"))
        self.assertEqual(another.get("X"), None)

        if os.path.exists(cache_file):
            os.remove(cache_file)

    def testGetBaseName(self):
        get_base_name_of_file("E:\\R\\a")
        self.assertEqual(get_base_name_of_file("./a"), "a")
        self.assertEqual(get_base_name_of_file("./a/"), "a")
        self.assertEqual(get_base_name_of_file("./a/c"), "c")
        self.assertEqual(get_base_name_of_file("./a//c"), "c")
        self.assertEqual(get_base_name_of_file("./a//b/c"), "c")
        self.assertEqual(get_base_name_of_file("./a//b/c/"), "c")

        self.assertEqual(get_base_name_of_file("E:\\R\\a"), "a")
        self.assertEqual(get_base_name_of_file("E:\\R\\a\\"), "a")
        self.assertEqual(get_base_name_of_file("E:\\R\\a\\c"), "c")
        self.assertEqual(get_base_name_of_file("E:\\R\\a\\c"), "c")
        self.assertEqual(get_base_name_of_file("E:\\R\\a\\b\\c"), "c")
        self.assertEqual(get_base_name_of_file("E:\\R\\a\\b\\c\\"), "c")

    def testPrettyFloat(self):
        self.assertEqual(get_pretty_float(1000.24575, count=1), "1E+3")
        self.assertEqual(get_pretty_float(1000.24575, count=2), "1.0E+3")
        self.assertEqual(get_pretty_float(1000.24575, count=3), "1.00E+3")
        self.assertEqual(get_pretty_float(1000.24575, count=4), "1000")
        self.assertEqual(get_pretty_float(1000.24575, count=5), "1000.2")
        self.assertEqual(get_pretty_float(1000.24575, count=6), "1000.25")
        self.assertEqual(get_pretty_float(1000.24575, count=7), "1000.246")
        self.assertEqual(get_pretty_float(1000.24575, count=8), "1000.2458")

    def testRandomChoice(self):
        rl = [i for i in range(10000)]

        x1 = random_choice(rl, k=100, unique=True)
        y1 = random_choice(rl, k=100, unique=True)

        self.assertTrue(len(set(x1)) == 100)
        self.assertTrue(len(set(y1)) == 100)
        self.assertFalse(";".join([str(i) for i in x1]) == ";".join([str(i) for i in y1]))



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import functools
import threading


class SingletonMixin(object):
    """
    thread safe singleton base class
    refer: https://gist.github.com/werediver/4396488

    # Based on tornado.ioloop.IOLoop.instance() approach.
    # See https://github.com/facebook/tornado
    """
    __singleton_lock = threading.Lock()
    __singleton_instance = None

    @classmethod
    def instance(cls):
        """

        :rtype: SingletonMixin
        """
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = cls()
        return cls.__singleton_instance


def singleton(func):
    """线程不安全的单例模式"""

    @functools.wraps(func)
    def wrapper(*args, **kw):
        if not hasattr(func, 'attr'):
            func.attr = func(*args, **kw)
        return func.attr

    return wrapper



# -*- coding:utf-8 -*-
from __future__ import absolute_import

from multiprocessing.pool import Pool

import cv2
import numpy as np
import os

from .image_utils import crop_image_array_by_boxes
from ..basic_tools.file_tools import list_files


def read_image(file_name):
    return cv2.imread(file_name, cv2.IMREAD_COLOR)


def save_image(file_name, image):
    cv2.imwrite(file_name, image)


def resize_image(image, width, height):
    return cv2.resize(image, (width, height), interpolation=cv2.INTER_CUBIC)


def convert_jpg_to_gray(src_file: str, dst_file: str):
    image = read_image(src_file)
    gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    cv2.imwrite(dst_file, gray_image)


def _convert_jpg_to_gray(args):
    src_file, dst_file = args
    convert_jpg_to_gray(src_file, dst_file)


def convert_jpg_to_gray_and_crop(src_file: str, dst_file: str, boxes: tuple):
    image = crop_image_array_by_boxes(read_image(src_file), boxes)
    gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    cv2.imwrite(dst_file, gray_image)
    # image = crop_image_by_boxes(get_image(src_file), boxes)
    # todo OpenCV Error: Assertion failed (scn == 3 || scn == 4) in cvtColor
    # gray_image = cv2.cvtColor(image_to_array(image), cv2.COLOR_BGR2GRAY)
    # cv2.imwrite(dst_file, gray_image)


def _convert_jpg_to_gray_and_crop(args):
    src_file, dst_file, boxes = args
    convert_jpg_to_gray_and_crop(src_file, dst_file, boxes)


def change_jpg_to_gray(src: str, dst: str, n_core: int = 1):
    if not os.path.exists(dst):
        os.mkdir(dst)

    args_list = [(full_name, os.path.join(dst, os.path.basename(full_name))) for full_name in list_files(src)]

    pool = Pool(n_core)
    pool.map_async(_convert_jpg_to_gray, [args for args in args_list if not os.path.exists(args[1])]).get()
    pool.close()


def change_jpg_to_gray_and_crop(src: str, dst: str, boxes_info: dict, n_core: int = 1):
    if not os.path.exists(dst):
        os.mkdir(dst)

    args_list = [
        (full_name, os.path.join(dst, os.path.basename(full_name)), boxes_info[os.path.basename(full_name)])
        for full_name in list_files(src)
    ]

    pool = Pool(n_core)
    pool.map_async(_convert_jpg_to_gray_and_crop, [args for args in args_list if not os.path.exists(args[1])]).get()
    pool.close()


def sharpen_image(image):
    image_blurred = cv2.GaussianBlur(image, (0, 0), 3)
    image_sharp = cv2.addWeighted(image, 1.5, image_blurred, -0.5, 0)
    return image_sharp


def create_mask(image):
    image_hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    sensitivity = 35
    lower_hsv = np.array([60 - sensitivity, 100, 50])
    upper_hsv = np.array([60 + sensitivity, 255, 255])

    mask = cv2.inRange(image_hsv, lower_hsv, upper_hsv)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (11, 11))
    mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel)

    return mask


def create_blank(width, height, rgb_color=(0, 0, 0)):
    """Create new image(numpy array) filled with certain color in RGB
        ref: https://stackoverflow.com/questions/4337902/how-to-fill-opencv-image-with-one-solid-color
        author: Kimmo

        usage:
            import cv2
            import numpy as np

            # Create new blank 300x300 red image
            width, height = 300, 300

            red = (255, 0, 0)
            image = create_blank(width, height, rgb_color=red)
            cv2.imwrite('red.jpg', image)
    """
    # Create black blank image
    image = np.zeros((height, width, 3), np.uint8)

    # Since OpenCV uses BGR, convert the color first
    color = tuple(reversed(rgb_color))
    # Fill image with color
    image[:] = color

    return image


__all__ = ("change_jpg_to_gray_and_crop", "read_image", "save_image", "resize_image", "convert_jpg_to_gray",
           "convert_jpg_to_gray_and_crop", "change_jpg_to_gray", "change_jpg_to_gray_and_crop",
           "sharpen_image", "create_mask", "create_blank")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import pickle

import faiss
import numpy as np

from pyxtools import global_init_logger, time_cost, random_choice

try:
    from pyxtools.faiss_tools import ImageIndexUtils
except ImportError:
    from pyxtools.pyxtools.faiss_tools import ImageIndexUtils


def load_simple_data(d: int, train_length: int, search_length: int) -> (np.ndarray, np.ndarray):
    train = np.random.random((train_length, d)).astype('float32')
    train[:, 0] += np.arange(train_length) / 1000.

    search = np.random.random((search_length, d)).astype('float32')
    search[:, 0] += np.arange(search_length) / 1000.

    return train, search


def load_feature_data(d: int, data_percent: float = 0.1) -> (np.ndarray, np.ndarray):
    cache = "/tmp/Feature/train.npy.pkl"
    with open(cache, "rb") as f:
        feature_list, _ = pickle.load(f)
        feature = np.vstack(feature_list).reshape((len(feature_list), d))

        # chosen data
        chosen_index_list = random_choice([index for index in range(len(feature_list))],
                                          k=int(len(feature_list) * data_percent), unique=True)
        search_feature_list = [feature_list[index] for index in chosen_index_list]

    return feature, np.vstack(search_feature_list).reshape((len(search_feature_list), d))


@time_cost
def test_search(train_feature: np.ndarray, search_feature: np.ndarray, d: int):
    index = faiss.index_factory(d, "Flat")
    index.add(train_feature)
    index.search(search_feature, k=5)


if __name__ == '__main__':
    global_init_logger()
    dimension = 1024
    train_data, search_data = load_feature_data(d=dimension, data_percent=0.01)
    logging.info("length of train data is {}, length of search data is {}".format(
        train_data.shape[0], search_data.shape[0]))
    test_search(train_feature=train_data, search_feature=search_data, d=dimension)



# Copyright (c) 2015-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD+Patents license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np

try:
    from pyxtools.faiss_tools import faiss
except ImportError:
    from pyxtools.pyxtools.faiss_tools import faiss

d = 64  # dimension
nb = 100000  # database size
nq = 10000  # nb of queries
np.random.seed(1234)  # make reproducible
xb = np.random.random((nb, d)).astype('float32')
xb[:, 0] += np.arange(nb) / 1000.
xq = np.random.random((nq, d)).astype('float32')
xq[:, 0] += np.arange(nq) / 1000.

res = faiss.StandardGpuResources()  # use a single GPU

# Using a flat index
index_flat = faiss.IndexFlatL2(d)  # build a flat (CPU) index

# make it a flat GPU index
gpu_index_flat = faiss.index_cpu_to_gpu(res, 0, index_flat)

gpu_index_flat.add(xb)  # add vectors to the index
print(gpu_index_flat.ntotal)

k = 4  # we want to see 4 nearest neighbors
distance_list, indices = gpu_index_flat.search(xq, k)  # actual search
print(indices[:5])  # neighbors of the 5 first queries
print(indices[-5:])  # neighbors of the 5 last queries

# Using an IVF index

nlist = 100
quantizer = faiss.IndexFlatL2(d)  # the other index
index_ivf = faiss.IndexIVFFlat(quantizer, d, nlist, faiss.METRIC_L2)
# here we specify METRIC_L2, by default it performs inner-product search

# make it an IVF GPU index
gpu_index_ivf = faiss.index_cpu_to_gpu(res, 0, index_ivf)

assert not gpu_index_ivf.is_trained
gpu_index_ivf.train(xb)  # add vectors to the index
assert gpu_index_ivf.is_trained

gpu_index_ivf.add(xb)  # add vectors to the index
print(gpu_index_ivf.ntotal)

k = 4  # we want to see 4 nearest neighbors
distance_list, indices = gpu_index_ivf.search(xq, k)  # actual search
print(indices[:5])  # neighbors of the 5 first queries
print(indices[-5:])  # neighbors of the 5 last queries



# -*- coding:utf-8 -*-
import logging
import pickle
import time
from threading import Lock

import faiss
import numpy as np
import os
from enum import Enum


class IndexType(Enum):
    accurate = 0
    fast = 1
    compress = 2
    auto = 3

    @property
    def to_train(self) -> bool:
        if self.name == "compress":
            return True
        return False

    @property
    def index_factory(self) -> str:
        if self.name == "accurate":
            return "Flat"
        elif self.name == "fast":
            return "IVFx,Flat"
        elif self.name == "compress":
            return "IVF100,PQ8"

        return "Flat"


class FaissManager(object):
    """
    """
    key_extend_list = "extend_list"
    key_class_id = "class_id"
    key_image_id = "index"
    not_found_id = -1

    def __init__(self, index_path: str, dimension: int,
                 index_type: IndexType = IndexType.accurate,
                 has_gpu: bool = False):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.faiss_index_file = index_path
        self.has_gpu = has_gpu
        self.faiss_index = None
        self.index_type = index_type
        self.dimension = dimension
        self._lock = Lock()

        # index info
        self._pkl_file = self.faiss_index_file + ".pkl"
        if os.path.exists(self._pkl_file):
            with open(self._pkl_file, "rb") as f:
                self.index_info_list = pickle.load(f)
        else:
            self.index_info_list = []

    @property
    def need_to_retrain(self) -> bool:
        if os.path.exists(self.faiss_index_file) and os.path.exists(self._pkl_file):
            return False

        if os.path.exists(self.faiss_index_file):
            os.remove(self.faiss_index_file)

        if os.path.exists(self._pkl_file):
            os.remove(self._pkl_file)

        return True

    @classmethod
    def parse_extend_list(cls, all_index_info: list) -> list:
        class_id_vs_images = {}
        for image_index, index_info in enumerate(all_index_info):
            class_id_vs_images.setdefault(index_info[cls.key_class_id], []).append(image_index)

        # add extend list
        for image_index, index_info in enumerate(all_index_info):
            extend_list_set = set(class_id_vs_images.get(index_info[cls.key_class_id], []))
            if image_index in extend_list_set:
                extend_list_set.remove(image_index)
            index_info[cls.key_extend_list] = list(extend_list_set)
        return all_index_info

    def train(self, feature_list, info_list: list):
        self.prepare_index()
        # extend_list
        self.index_info_list = self.parse_extend_list(info_list)

        feature = self.reshape_feature_list(feature_list)

        if self.index_type.to_train:
            self.logger.info("training index...")
            time_start = time.time()
            self.faiss_index.train(feature)  # nb * d
            self.logger.info("success to train index! Cost {} seconds!".format(time.time() - time_start))

        self.faiss_index.add(feature)
        self.save()

    def reshape_feature_list(self, feature_list) -> np.ndarray:
        """  """
        feature = feature_list
        if isinstance(feature_list, list):
            # feature_list shape: [(1, self.d), (1, self.d)]
            feature = np.vstack(feature_list).reshape((len(feature_list), self.dimension))
        return feature

    def search(self, feature_list, top_k=10) -> (np.ndarray, np.ndarray):
        self.prepare_index()
        feature_list = self.reshape_feature_list(feature_list)
        length = feature_list.shape[0]

        time_start = time.time()
        distance_list, indices = self.faiss_index.search(feature_list, top_k)
        self.logger.info("cost {}s to search {} feature".format(time.time() - time_start, length))

        distance_list = distance_list.reshape((length, top_k))
        indices = indices.reshape((length, top_k))

        return distance_list, indices

    def _restore(self):
        """

        :rtype: object
        """
        if not os.path.exists(self.faiss_index_file):
            raise Exception("{} not exists!".format(self.faiss_index_file))

        return faiss.read_index(self.faiss_index_file)

    def prepare_index(self):
        # index
        if self.faiss_index is not None:
            return

        if os.path.exists(self.faiss_index_file):
            self.faiss_index = self._restore()
            return

        # create index
        if self.index_type.index_factory:
            self.faiss_index = faiss.index_factory(self.dimension, self.index_type.index_factory)
            if self.has_gpu:
                res = faiss.StandardGpuResources()  # use a single GPU
                self.faiss_index = faiss.index_cpu_to_gpu(res, 0, self.faiss_index)

    def save(self, ):
        assert self.faiss_index is not None

        with self._lock:
            faiss.write_index(self.faiss_index, self.faiss_index_file)

            with open(self._pkl_file, "wb") as f:
                pickle.dump(self.index_info_list, f)


__all__ = ("faiss", "IndexType", "FaissManager")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import unittest

import numpy as np
import os
import random
import shutil

try:
    from pyxtools.faiss_tools import ImageIndexUtils
except ImportError:
    from pyxtools.pyxtools.faiss_tools import ImageIndexUtils


class TestImageIndexUtils(unittest.TestCase):
    def setUp(self):
        self.index_path = "./data"
        if os.path.exists(self.index_path):
            shutil.rmtree(self.index_path)

        self.dimension = 10
        self.image_utils = ImageIndexUtils(index_dir=self.index_path, dimension=self.dimension)

    def tearDown(self):
        shutil.rmtree(self.index_path)
        self.assertFalse(os.path.exists(self.index_path))

    def test_cls(self):
        # prepare index
        count = 20
        feature_list = [
            np.asarray([i * 0.1 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension))
        ]
        image_info_list = [{"i": 0, "j": 10, "class_id": 1}]
        for i in range(count - 1):
            feature_list.append(
                np.asarray([i * random.random() for i in range(self.dimension)], dtype=np.float32)
                    .reshape((1, self.dimension))
            )
            image_info_list.append({"i": i + 1, "j": 11 + i, "class_id": i + 1})

        self.image_utils.add_images(image_feature_list=feature_list,
                                    image_info_list=image_info_list)
        self.assertTrue(os.path.exists(self.image_utils.manager.faiss_index_file))
        self.assertEqual(count, len(self.image_utils.manager.index_info))

        # search: top 1
        top_k = 1
        feature_list = [
            np.asarray([i * 0.1 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension))
        ]
        result_list = self.image_utils.image_search(feature_list=feature_list, top_k=top_k)
        self.assertEqual(len(feature_list), len(result_list))
        print("result for {} is :\n{}".format(feature_list, result_list))

        feature_list = [
            np.asarray([i * 0.1 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension)),
            np.asarray([i * 0.2 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension)),
        ]
        result_list = self.image_utils.image_search(feature_list=feature_list, top_k=top_k)
        self.assertEqual(len(feature_list), len(result_list))
        print("result for {} is :\n{}".format(feature_list, result_list))

        # search: top 5
        top_k = 5
        feature_list = [
            np.asarray([i * 0.1 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension))
        ]
        result_list = self.image_utils.image_search(feature_list=feature_list, top_k=top_k)
        self.assertEqual(len(feature_list), len(result_list))
        print("result for {} is :\n{}".format(feature_list, result_list))

        feature_list = [
            np.asarray([i * 0.1 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension)),
            np.asarray([i * 0.2 for i in range(self.dimension)], dtype=np.float32).reshape((1, self.dimension)),
        ]
        result_list = self.image_utils.image_search(feature_list=feature_list, top_k=top_k)
        self.assertEqual(len(feature_list), len(result_list))
        print("result for {} is :\n{}".format(feature_list, result_list))



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
from urllib.parse import urlparse

import requests

logger = logging.getLogger(__name__)


def simple_request_get(url):
    return requests.get(
        url,
        headers={
            'user-agent': 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.2 (KHTML, like Gecko)' +
                          ' Chrome/22.0.1216.0 Safari/537.2'
        },
        timeout=15).content


def proxy_request_get(url, proxy_addr, proxy_port):
    socks_url = "socks5://{}:{}".format(proxy_addr, proxy_port)
    proxies = {'http': socks_url, 'https': socks_url}
    return requests.get(
        url,
        headers={
            'user-agent': 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.2 (KHTML, like Gecko)' +
                          ' Chrome/22.0.1216.0 Safari/537.2'
        },
        proxies=proxies,
        timeout=15).content


def request_get_by_web_api(url, ):
    default_headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36",
        "DNT": "1",
    }

    def get_response_from_codebeautify(_url):
        """
        https://codebeautify.org/source-code-viewer
        :param _url:
        :return:
        """
        headers = dict(default_headers)
        headers.update({
            "Accept": "text/plain, */*; q=0.01",
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Origin": "https://codebeautify.org",
            "Referer": "https://codebeautify.org/source-code-viewer",
        })
        return requests.post(
            "https://codebeautify.com/URLService",
            data={"path": _url},
            headers=headers,
            timeout=15).content

    return get_response_from_codebeautify(url)


def strong_request_get(url, proxy_addr="127.0.0.1", proxy_port=1080):
    try:
        return simple_request_get(url)
    except Exception as e:
        logger.error(e)
    try:
        return proxy_request_get(url, proxy_addr, proxy_port)
    except Exception as e:
        logger.error(e)
        return request_get_by_web_api(url)


def get_domain(url):
    """

    :rtype: str
    :type url: str
    """
    parsed_uri = urlparse(url)
    return '{uri.scheme}://{uri.netloc}/'.format(uri=parsed_uri)


__all__ = ("strong_request_get", "simple_request_get", "proxy_request_get", "get_domain", "request_get_by_web_api")



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import math
import uuid

import hashlib


def create_guid() -> str:
    return str(uuid.uuid1()).lower()


def create_fake_random_string(length: int) -> str:
    assert length > 0

    md5_list = [get_md5(create_guid().encode("utf-8")) for _ in range(math.ceil(length / 30) + 1)]

    return "".join(md5_list)[:length]


def get_md5(string) -> str:
    """
    use in python3.6:
    """
    if isinstance(string, str):
        hash_md5 = hashlib.md5(string.encode("utf-8"))
    else:
        hash_md5 = hashlib.md5(string)
    return hash_md5.hexdigest()


def get_md5_for_file(file_or_filename) -> str:
    if isinstance(file_or_filename, str):
        hash_md5 = hashlib.md5()
        with open(file_or_filename, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: file_or_filename.read(4096), b""):
        hash_md5.update(chunk)
    return hash_md5.hexdigest()



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import fnmatch
import json
import logging
import os
import subprocess
import sys
from optparse import OptionParser
from subprocess import check_output

import requests

from .crlf_tools import change_crlf_mode_of_file, CRLF_VALUE
from ..basic_tools.log import global_init_logger

logger = logging.getLogger(__name__)

"""
git config:
```
[core]
    fileMode = false
    autocrlf = true
    safecrlf = true
```

create .gitattributes file with content:
```
# Set the default behavior, in case people don't have core.autocrlf set.
* text eol=lf

core.autocrlf=true
core.fileMode=false
```

fore change crlf if found error before `git commit`
```
 python -c "from pyxtools import git_crlf_helper as g;g()" -d . -t lf -i *.py -e *.pyc
```

"""


def git_list_status_file(path) -> list:
    """ list all file showed in `cd path && git status -s` """
    raw_path = os.getcwd()
    try:
        os.chdir(path)
        raw_lines = check_output(["git", "status", "-s"]).decode("utf-8").split("\n")
        logger.info("result of `git status -s` is {}".format("\n".join(raw_lines)))

        def _get_file_name(_line: str) -> str:
            blank_index = _line.find(" ")
            if blank_index > -1:
                return _line[blank_index:]
            return ""

        lines = [_get_file_name(clear_line) for clear_line in [line.strip("\r").strip() for line in raw_lines]]
        file_list = [os.path.join(path, file_name) for
                     file_name in [line.strip() for line in lines] if len(file_name) > 0]
        logger.info("result of `git_list_status_file` is ['{}']".format("'\n'".join(file_list)))
        return file_list
    finally:
        os.chdir(raw_path)


def git_crlf_fatal_file(path) -> list:
    """ list all file showed in `cd path && git status -s` """
    raw_path = os.getcwd()
    try:
        os.chdir(path)
        try:
            output = check_output(["git", "add", "."], shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            output = e.output

        raw_lines = output.decode("utf-8").split("\n")
        logger.info("result of `git add .` is {}".format("\n".join(raw_lines)))
        fatal_line = raw_lines[0].strip()

        file_list = []
        if fatal_line.find("fatal") > -1:
            file_name = os.path.join(path, fatal_line.split(" ")[-1].rstrip("."))
            if os.path.exists(file_name) and os.path.isfile(file_name):
                file_list = [file_name]
            else:
                logger.warning("{} not exists!".format(file_name))
        if file_list:
            logger.info("result of `git_crlf_fatal_file` is ['{}']".format("'\n'".join(file_list)))
        else:
            logger.info("result of `git_crlf_fatal_file` is []")
        return file_list
    finally:
        os.chdir(raw_path)


def is_git_repo_log_dir(log_dir: str) -> bool:
    return os.path.isdir(log_dir) and \
           os.path.exists(os.path.join(log_dir, "index")) and \
           os.path.exists(os.path.join(log_dir, "HEAD"))


def find_git_log_dir(path: str) -> (str, str):
    while True:
        if not os.path.exists(path) or not os.path.isdir(path):
            raise ValueError("fail to find git log dir")

        git_log = os.path.join(path, ".git")
        if os.path.exists(git_log):
            if is_git_repo_log_dir(git_log):
                return path, git_log
        else:
            with open(git_log, "r", encoding="utf-8") as f:
                for line in f:
                    if line.find("gitdir:") > -1:
                        relative_path = line[line.find("gitdir:"):].strip("\n").strip("\r").strip()
                        real_git_log = os.path.join(path, relative_path)
                        if is_git_repo_log_dir(real_git_log):
                            return path, real_git_log

        path = os.path.join(path, "..")


def git_crlf_helper():
    global_init_logger()
    hstr = '%prog custom help string'
    parser = OptionParser(hstr, description='custom description', version='%prog 1.0')
    parser.add_option('-d', '--dir', action='store', dest='dir', default=".", help='working dir')
    parser.add_option('-i', '--include', action='store', dest='include', default="",
                      help='include pattern, join by `,`')
    parser.add_option('-e', '--exclude', action='store', dest='exclude', default="",
                      help='exclude pattern, join by `,`')
    parser.add_option('-t', '--target', action='store', dest='target', default="lf",
                      help='target: crlf, lf, cr')
    if sys.argv[0] == "-c":
        argv = list(sys.argv)
        argv[0] = __file__
        options, args = parser.parse_args(argv)
    else:
        options, args = parser.parse_args(sys.argv)

    if options.target in CRLF_VALUE:
        options.target = CRLF_VALUE[options.target]

    include = None
    if len(options.include) > 0:
        include = [pattern for pattern in options.include.split(",") if len(pattern) > 0]

    file_list = git_list_status_file(options.dir)
    all_file_list = []
    for file_name in file_list:
        if not os.path.exists(file_name):
            logger.warning("{} not exists, ignore it".format(file_name))
        else:
            all_file_list.append(file_name)
    file_list = list(all_file_list)

    if include is not None and len(include) > 0:
        all_file_list = []
        for pattern in include:
            all_file_list.extend(fnmatch.filter(file_list, pattern))
        file_list = list(set(all_file_list))

    exclude = None
    if len(options.exclude) > 0:
        exclude = [pattern for pattern in options.exclude.split(",") if len(pattern) > 0]
    if exclude is not None and len(exclude) > 0:
        exclude_file_list = []
        for pattern in exclude:
            exclude_file_list.extend(fnmatch.filter(file_list, pattern))
        file_list = list(set(file_list) - set(exclude_file_list))

    if len(file_list) == 0:
        logger.info("not file in unstage status!")
        return

    [change_crlf_mode_of_file(file_name, options.target) for file_name in file_list]

    count = 100
    file_list = git_crlf_fatal_file(options.dir)
    while count > 0 and len(file_list) > 0:
        change_crlf_mode_of_file(file_list[0], options.target)
        file_list = git_crlf_fatal_file(options.dir)
        count -= 1


def md2html_by_github(content: str) -> str:
    """ use github api """
    logger.info("requesting github...")
    url = "https://api.github.com/markdown"
    return requests.post(url, data=json.dumps({"text": content, "mode": "markdown"})).text


def git_export(export_zip_file_name: str, git_project_path: str = "./"):
    """
        导出git项目的纯净代码

        git archive --format zip --output "./output.zip" master -0
        # 将代码导出并 zip 打包后放在当前目录下，`output.zip`就是需要的文件，`-0`的意思是不压缩
    """
    current = os.getcwd()
    try:
        os.chdir(git_project_path)
        process = subprocess.Popen(
            ["git", "archive", "master", "--format", "zip", "--output", export_zip_file_name, ]
        )
        process.wait()
    finally:
        os.chdir(current)


def git_fetch_all(git_path: str):
    """ run `cd git_path && git fetch --all` """
    raw_path = os.getcwd()
    try:
        os.chdir(git_path)
        subprocess.Popen(["git", "fetch", "--all"]).wait()
    finally:
        os.chdir(raw_path)


__all__ = ("git_list_status_file", "is_git_repo_log_dir", "git_crlf_fatal_file",
           "find_git_log_dir", "md2html_by_github", "git_crlf_helper", "git_export",
           "git_fetch_all")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import math

import PIL.Image as Image
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import os
from matplotlib.colors import LogNorm


def get_image(path: str, width: int = None, height: int = None) -> Image:
    img = Image.open(path)
    if width is None and height is None:
        return img

    return img.resize((width, height), Image.ANTIALIAS)


def create_blank_image(image: Image, mode: str = "RGB") -> Image:
    if mode == "L":
        blank_image = image.copy().convert('L')
        blank_image_array = image_to_array(blank_image).copy()
        for i in range(len(blank_image_array)):
            for j in range(len(blank_image_array[0])):
                blank_image_array[i][j] = 0
        return np_image_to_pil_image(blank_image_array, mode="L")
    elif mode == "RGB":
        blank_image = image.copy()
        blank_image_array = image_to_array(blank_image).copy()
        for i in range(len(blank_image_array)):
            for j in range(len(blank_image_array[0])):
                blank_image_array[i][j][:] = 255
        return np_image_to_pil_image(blank_image_array)


def show_images(image_list, mode=None, image_save_file: str = None, fig_size=(64, 64), dpi=None):
    """

    :param mode: None|str, for example, Greys
    :param image_list: Image object list
    :param image_save_file: plt.show() if image_save_file is None, else save image
    :param fig_size: tuple
    :param dpi: int,
    """
    if fig_size is None:
        fig = plt.figure(dpi=dpi)
    else:
        fig = plt.figure(figsize=fig_size, dpi=dpi)

    index = 0
    rows = len(image_list)
    columns = len(image_list[0])
    for row_image in image_list:
        blank_image = None
        for image in row_image:
            index += 1
            ax = fig.add_subplot(rows, columns, index)
            ax.axis('off')
            if image is None:
                if blank_image is None:
                    blank_image = create_blank_image(row_image[0])
                image = blank_image

            if mode is None:
                plt.imshow(image)
            else:
                plt.imshow(image, cmap=mode)

    plt.axis('off')
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)

    if image_save_file is None:
        plt.show()
    else:
        plt.savefig(image_save_file)


def show_images_file(image_list, mode=None, image_save_file=None, image_size: tuple = None, dpi: int = None):
    """

    :param mode: None|str, for example, Greys
    :param image_list: image file name list
    :param image_save_file: plt.show() if image_save_file is None, else save image
    :param image_size: (width, height)
    :param dpi: int
    """
    image_obj_list = []
    for row_image in image_list:
        row_image_obj_list = []
        for image_file in row_image:
            if os.path.exists(image_file):
                if image_size:
                    img = get_image(image_file, image_size[0], image_size[1])
                else:
                    img = get_image(image_file)
                row_image_obj_list.append(img)
            else:
                row_image_obj_list.append(None)
        image_obj_list.append(list(row_image_obj_list))

    if image_size:
        if dpi is None:
            dpi = 100 if image_size[0] > 100 else image_size[0]

        show_images(
            image_obj_list, mode, image_save_file, dpi=dpi,
            fig_size=(math.ceil(image_size[0] * max([len(row) for row in image_list]) / dpi),
                      math.ceil(image_size[1] * len(image_list) / dpi))
        )
    else:
        show_images(image_obj_list, mode, image_save_file, dpi=dpi)


def show_image(image: Image, mode: str = None, text: str = None, ):
    """

    :type text: None|str
    :param mode: None|str, for example, gray, Greys
    :param image: Image object
    """

    f = plt.figure()
    ax = f.add_subplot(111)
    if text:
        ax.text(0.1, 0.9, text, ha='center', va='center', transform=ax.transAxes)

    if mode:
        plt.imshow(image, cmap=mode)
    else:
        plt.imshow(image)


def np_image_to_pil_image(image_array: np.ndarray, mode="RGB"):
    """

    :rtype: Image
    :type mode: str
    :type image_array: np.array, (h, w, c)
    """
    return Image.fromarray(image_array.astype(np.int8), mode)


def image_to_array(image: Image, shape=None):
    """
    width x height
    array: shape (height, width)
    shape 0 y, height; shape 1 x, width
    :type shape: None|tuple
    :type image: Image
    """
    array = np.asarray(image)

    if shape:
        return array.reshape(shape)

    return array


def show_hot_graph(xyz_list, image_save_file=None):
    x, y, z = xyz_list
    x_min = np.min(x)
    x_max = np.max(x)
    y_min = np.min(y)
    y_max = np.max(y)
    # z_min = np.min(z)
    # z_max = np.max(z)

    height = y_max - y_min + 1
    width = x_max - x_min + 1
    arr = np.zeros((height, width))  # arr 热力图中的值阵

    for i in range(len(x)):
        # arr[y[i] - y_min, x[i] - x_min] = (z[i] - z_min) * 10 / (z_max - z_min) + 1
        arr[y[i] - y_min, x[i] - x_min] = z[i]

    # 热力图默认左上为0,0
    # 所以热力图的显示和arr是一致的
    # 未解决以左下为0,0,
    plt.imshow(arr, extent=(np.amin(x), np.amax(x), np.amax(y), np.amin(y)),
               cmap=cm.hot, norm=LogNorm())
    plt.colorbar()

    # plt.figure(figsize=(width, height)) # todo error

    if image_save_file is None:
        plt.show()
    else:
        plt.savefig(image_save_file)


def crop_image_by_boxes(image: Image, boxes: tuple) -> Image:
    """

    :param image: Image
    :param boxes: (x0, y0, x1, y1) == (left, upper, right, lower); upper < lower, left < right
    :return: Image
    """
    # left = 2407
    # top = 804
    # width = 300
    # height = 200
    # box = (left, top, left + width, top + height)
    return image.crop(boxes)


def crop_image_array_by_boxes(image: np.ndarray, boxes: tuple) -> np.ndarray:
    """

    :param image: Image
    :param boxes: (x0, y0, x1, y1) == (left, upper, right, lower); upper < lower, left < right
    :return: Image
    """
    return image[boxes[1]:boxes[3], boxes[0]:boxes[2], :]


def gen_tf_bounding_boxes(boxes):
    """
        change (left, upper, right, lower) -> (offset_height, offset_width, target_height, target_width)
        Note: upper < lower, left < right
    """
    left, upper, right, lower = boxes
    offset_height = upper
    offset_width = left
    target_height = lower - upper
    target_width = right - left
    return offset_height, offset_width, target_height, target_width


def expend_bounding_box(boxes, pix: int = 5):
    left, upper, right, lower = boxes
    return max(0, left - pix), max(0, upper - pix), right + pix, lower + pix


__all__ = ("get_image", "create_blank_image", "show_images", "show_images_file", "show_image",
           "np_image_to_pil_image", "image_to_array", "show_hot_graph", "crop_image_by_boxes",
           "crop_image_array_by_boxes", "gen_tf_bounding_boxes", "expend_bounding_box")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import socket

import requests

from .async_http_tool import async_request_get


def port_is_used(addr: str, port: int) -> bool:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((addr, int(port)))
            return True
    except Exception as e:
        logging.error(e)

    return False


def proxy_is_valid(proxy_url, url_tuple=None, timeout=10, async_mode=False):
    """

    :type async_mode: bool
    :param proxy_url: str, like "socks5://127.0.0.1:1080"
    :rtype: bool
    :type timeout: int
    :type url_tuple: None|tuple
    """
    if url_tuple is None:
        url_tuple = (
            "https://www.baidu.com/robots.txt",
            "https://www.qq.com/robots.txt",
            "http://www.hao123.com/robots.txt"
        )

    if len(url_tuple) == 1:
        async_mode = False

    if async_mode is False:
        # blocking mode
        proxies = {'http': proxy_url, 'https': proxy_url}

        def is_work(_url):
            try:
                response = requests.get(_url, proxies=proxies, timeout=timeout)
                return len(response.content) > 0
            except Exception as e:
                logging.error("Failed to connect to {}: {}".format(_url, e))
                return False

        result_list = [is_work(url) for url in url_tuple]

    else:
        # async_mode mode
        def response_func(content):
            return len(content) > 0

        raw_list = async_request_get(
            list(url_tuple), time_out=timeout, proxy=proxy_url, sema_count=5, response_callback=response_func
        )
        result_list = []
        for result in raw_list:
            if isinstance(result, bool):
                result_list.append(result)
            else:
                result_list.append(False)

    return result_list.count(True) >= len(result_list) / 2.0


def get_random_unopen_port(local_addr: str) -> int:
    """

    :param local_addr: ip to bind, 127.0.0.1
    :return: unopen port
    """
    with socket.socket() as sock:
        sock.bind((local_addr, 0))
        ip, port = sock.getsockname()
        return port



# -*- coding:utf-8 -*-
from __future__ import absolute_import

"""
warning: this package not add to top level package now

"""
import json
import datetime
import logging
import operator
import os
import re
from itertools import takewhile
from ..dev_tools.git_tools import md2html_by_github
from ..basic_tools.md5_tools import get_md5
from ..basic_tools.cache_tools import SqliteCache

from flask_flatpages.page import Page as OldPage, cached_property, yaml
from werkzeug.utils import import_string

logger = logging.getLogger(__name__)


class UrlTemplate(object):
    @classmethod
    def decode(cls, path):
        """ file path to url"""
        if not path.endswith(".html"):
            path += ".html"
        return path

    @classmethod
    def encode(cls, url):
        """url to file path"""
        if url.endswith(".html"):
            return url[:-5]
        return url


class HtmlCache(object):
    cache = SqliteCache("./cache.dat")

    @classmethod
    def get_content_id_and_html(cls, md_id):
        """

        :type md_id: str
        :rtype: (str, str)
        """
        result = cls.cache.get(md_id)
        if result is None:
            return None, None
        return result[0], result[1]

    @classmethod
    def set_content_id_and_html(cls, md_id, content_id, html_value):
        cls.cache.set(key=md_id, value=(content_id, html_value))


class Page(OldPage):
    def __init__(self, path, meta, body, html_renderer):
        """

        :type body: str
        :type meta: str
        :type path: str
        """
        super(Page, self).__init__(UrlTemplate.decode(path), meta, body, html_renderer)

        # date re
        self.date_re = re.compile(r"[\d]{4}-[\d]{2}-[\d]{2}(?=-)")

        # 执行meta方法
        if self.meta:
            pass

    @cached_property
    def html(self):
        """The content of the page, rendered as HTML by the configured
        renderer.
        """
        file_name_id = get_md5(self.path.encode("utf-8"))

        old_content_id, content_html = HtmlCache.get_content_id_and_html(file_name_id)
        current_content_id = get_md5(self.body.encode("utf-8"))

        if content_html is None or old_content_id != current_content_id:
            content_html = None
            try:
                content_html = md2html_by_github(self.body)
                msg = json.loads(content_html)
                logger.warning("get msg from github: {}".format(msg))
            except Exception as e:
                logger.error(e)
                if content_html is not None:
                    HtmlCache.set_content_id_and_html(file_name_id, current_content_id, content_html)

        return content_html

    @cached_property
    def meta(self):
        """A dict of metadata parsed as YAML from the header of the file.
        """
        meta = yaml.safe_load(self._meta)
        # YAML documents can be any type but we want a dict
        # eg. yaml.safe_load('') -> None
        #     yaml.safe_load('- 1\n- a') -> [1, 'a']
        if not meta:
            if self._meta:
                # 还原body
                self.body = "\n".join((self._meta, self.body))

            meta = {}

        if not isinstance(meta, dict):
            raise ValueError("Excpected a dict in metadata for '{0}', got {1}".
                             format(self.path, type(meta).__name__))

        # title
        if "title" not in meta:
            _title = None
            if self._meta:
                _title = self._meta.replace("#", "").strip()

            if _title is None:
                _title = os.path.basename(self.path)

            meta["title"] = _title or "日志"

        # date
        if "date" not in meta:
            meta["date"] = self._get_date_from_path(self.path) or datetime.date.today()

        return meta

    def _get_date_from_path(self, path):
        """
            文件命名: 2018-01-31-xxx.md
        :rtype: datetime.date
        """
        result = self.date_re.findall(path)
        if result:
            return datetime.datetime.strptime(result[0], "%Y-%m-%d").date()
        return None


def flat_pages_parse(self, content, path):
    """Parse a flatpage file, i.e. read and parse its meta data and body.

    :return: initialized :class:`Page` instance.
    """
    lines = iter(content.split('\n'))

    # Read lines until an empty line is encountered.
    meta = '\n'.join(takewhile(operator.methodcaller('strip'), lines))
    # The rest is the content. `lines` is an iterator so it continues
    # where `itertools.takewhile` left it.
    content = '\n'.join(lines)

    # Now we ready to get HTML renderer function
    html_renderer = self.config('html_renderer')

    # If function is not callable yet, import it
    if not callable(html_renderer):
        html_renderer = import_string(html_renderer)

    # Make able to pass custom arguments to renderer function
    html_renderer = self._smart_html_renderer(html_renderer)

    # Initialize and return Page instance
    return Page(path, meta, content, html_renderer)


__all__ = ("flat_pages_parse", "Page", "HtmlCache", "UrlTemplate")



# -*- coding:utf-8 -*-
import logging

import os

from .faiss_utils import FaissManager


class ImageIndexUtils(object):
    key_extend_list = "extend_list"

    def __init__(self, index_dir: str, dimension: int):
        self.logger = logging.getLogger(self.__class__.__name__)
        if not os.path.exists(index_dir):
            os.mkdir(index_dir)
        self.db_index_dir = index_dir
        self.manager = FaissManager(
            index_path=os.path.join(index_dir, "faiss.index"),
            dimension=dimension
        )
        self.dimension = dimension
        self._key_distance = "distance"
        self._key_top_k = "top"

    def image_search(self, feature_list: list, top_k: int = 3, extend: bool = False) -> list:
        """
            for each feature, return list like: [(1, [100], 0.10), (3,[], 0.30), ...]
        """
        return [
            result for result in self.image_search_iterator(feature_list=feature_list, top_k=top_k, extend=extend)
        ]

    def get_image_info(self, image_index: int) -> dict:
        return self.manager.index_info_list[image_index]

    def image_search_iterator(self, feature_list: list, top_k: int = 3, extend: bool = False):
        """
            for each feature, return list like: [(1, [100], 0.10), (3,[], 0.30), ...]
        """
        distance_list, indices = self.manager.search(feature_list, top_k=top_k)

        for index in range(distance_list.shape[0]):
            image_result_list = []

            for i in range(top_k):
                image_index = indices[index][i]
                if image_index == self.manager.not_found_id:
                    break

                if extend:
                    image_result_list.append(
                        (image_index,
                         self.manager.index_info_list[image_index].get(self.manager.key_extend_list, []),
                         distance_list[index][i])
                    )
                else:
                    image_result_list.append((image_index, [], distance_list[index][i]))

            yield image_result_list

    def add_images(self, image_feature_list: list, image_info_list: list):
        assert len(image_feature_list) == len(image_info_list)
        self.manager.train(image_feature_list, info_list=image_info_list)


__all__ = ("ImageIndexUtils",)



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import shutil
import sys
import time
import traceback
from subprocess import Popen, PIPE, STDOUT, check_output


def backup_linux_history(target_file_path):
    def list_history_lines():
        cmd = Popen("bash -i -c  'history -r;history' ", shell=True, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
        cmd.wait()
        return [_line.decode("utf-8") for _line in cmd.stdout.readlines()]

    def can_connect(first_list, second_list):
        if not first_list or len(second_list) < len(first_list):
            return False

        for _index, _value in enumerate(first_list):
            if second_list[_index] != _value:
                return False

        return True

    def get_start_connect_line_index(history_line_list, ):
        # total count
        _total_count = 0
        with open(target_file_path, "r", encoding="utf-8") as f:
            for _ in f:
                _total_count += 1

        # start checking line index
        _history_line_count = len(history_line_list)
        if _total_count <= _history_line_count:
            _start_checking_line_index = 0
        else:
            _start_checking_line_index = _total_count - _history_line_count

        # calc
        _line_index = 0
        with open(target_file_path, "r", encoding="utf-8") as f:
            # skip not checking lines
            if _start_checking_line_index > 0:
                for _ in f:
                    _line_index += 1
                    if _line_index >= _start_checking_line_index:
                        break

            _target_checking_lines = [" ".join(_line.lstrip(" ").rstrip("\n").split(" ")[1:]) for _line in f]

        _history_checking_lines = [" ".join(_line.lstrip(" ").rstrip("\n").split(" ")[1:]) for _line in
                                   history_line_list]

        _current_total_index = _line_index
        for _index, _line in enumerate(_target_checking_lines):
            _current_total_index = _index + _line_index
            if can_connect(_target_checking_lines[_index:], _history_checking_lines):
                return _current_total_index

        return _current_total_index

    history_lines = list_history_lines()
    start_connect_line_index = get_start_connect_line_index(history_lines)
    tmp_target_file_path = target_file_path + ".{}.tmp".format(int(time.time()))

    with open(tmp_target_file_path, "w", encoding="utf-8") as fw:
        with open(target_file_path, "r", encoding="utf-8") as fr:
            for line_index, line in enumerate(fr):
                if line_index >= start_connect_line_index:
                    break
                fw.write(line)

        for line in history_lines:
            fw.write(line)

    shutil.move(tmp_target_file_path, target_file_path)


def stacktraces():
    """
    ref: http://san-yun.iteye.com/blog/1639763
    :return:
    """
    code = []
    for threadId, stack in sys._current_frames().items():
        code.append("# ThreadID: %s" % threadId)
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                code.append("  %s" % (line.strip()))

    return "\n".join(code)


def windows_process_exists_by_name(process_name: str) -> bool:
    """
    ref: https://stackoverflow.com/questions/7787120/python-check-if-a-process-is-running-or-not
    :rtype: bool
    :param process_name: str
    """
    # shell=True hides the shell window, stdout to PIPE enables
    raw_lines = check_output(
        ['TASKLIST', '/FI', 'imagename eq {}'.format(process_name)], shell=True) \
        .decode("utf-8").split("\r\n")

    # if TASKLIST returns single line without processname: it's not running
    return len(raw_lines) > 1 and process_name in raw_lines[-1]



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import time

import numpy as np
from enum import Enum

_np_zero = 1e-15


class NormType(Enum):
    """
    feature = np.random.random((3, 10))
    assert feature.shape == (3, 10)

    if norm_type == NormType.l2:
        normalized_feature = normalize(feature, norm='l2', axis=0)  # 对d(10)维分别l2正则化
        # normalized_feature = feature/np.linalg.norm(feature, ord=2, axis=0)
    elif norm_type == NormType.all:
        normalized_feature = feature / np.sqrt(np.sum(np.power(feature, 2)))
    else:
        normalized_feature = feature
    """
    none = 0
    l2 = 1
    all = 2

    def normalize(self, feature: np.ndarray) -> np.ndarray:
        normalize_feature, _ = self.normalize_and_return_norm(feature)
        return normalize_feature

    def normalize_and_return_norm(self, feature: np.ndarray) -> (np.ndarray, object):
        if self.name == "all":
            _norm_float = np.sqrt(np.sum(np.power(feature, 2)))  # _norm_float is float
            if _norm_float == 0.0:
                logging.warning("using {} because norm == 0.0 when normalize with NormType.all mode".format(_np_zero))
                _norm_float += _np_zero
            return feature / _norm_float, float(_norm_float)
        elif self.name == "l2":
            logging.warning("using {} if norm == 0 when normalize with NormType.l2 mode".format(_np_zero))
            norm_feature = np.linalg.norm(feature, ord=2, axis=0)
            safe_norm_feature = np.add(norm_feature, _np_zero, out=norm_feature, where=norm_feature == 0)
            return feature / safe_norm_feature, safe_norm_feature
        elif self.name == "none":
            return feature, None
        else:
            raise ValueError("unknown NormType: {}".format(self.name))


def calc_distance_pairs(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
    """
    Args:
        vec1: array, shape is [m, d]
        vec2: array, shape is [n, d]

    Returns:
        array: shape is [m, n]
    """
    assert vec1.shape[1] == vec2.shape[1]
    assert len(vec1.shape) == len(vec2.shape) == 2

    distance_matrix = np.sqrt(
        -2 * np.dot(vec1, vec2.T) +
        np.sum(np.square(vec2), axis=1) +
        np.transpose([np.sum(np.square(vec1), axis=1)])
    )

    nan_where = np.argwhere(np.isnan(distance_matrix))
    if len(nan_where):
        logging.debug("found nan_where {}".format(nan_where))
        for index in nan_where:
            _new_distance = np.linalg.norm(vec1[index[0]] - vec2[index[1]], ord=2, axis=0)
            logging.debug("distance matrix [{}] is {}, new distance is {}".format(
                index, distance_matrix[index[0], index[1]], _new_distance))
            distance_matrix[index[0], index[1]] = _new_distance

    return distance_matrix


def _calc_distance_pairs_by_scipy(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
    """

    Args:
        vec1: array, shape is [m, d]
        vec2: array, shape is [n, d]

    Returns:
        array: shape is [m, n]
    """
    assert vec1.shape[1] == vec2.shape[1]
    assert len(vec1.shape) == len(vec2.shape) == 2
    from scipy import spatial

    return spatial.distance_matrix(vec1, vec2, p=2, threshold=1e12)


def _test_benchmark_distance():
    m, n, d = 10000, 20000, 512
    a = np.random.random((m, d))
    b = np.random.random((n, d))

    # np distance
    _time_start = time.time()
    dis = calc_distance_pairs(a, b)
    assert dis.shape == (m, n)
    print("time cost {}s".format(time.time() - _time_start))

    # scipy distance: OOM
    _time_start = time.time()
    dis = _calc_distance_pairs_by_scipy(a, b)  # oom
    assert dis.shape == (m, n)
    print("time cost {}s".format(time.time() - _time_start))


__all__ = ("NormType", "calc_distance_pairs",)



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import os
import shutil
import sys
from optparse import OptionParser

from ..basic_tools.file_tools import compress_by_tar, list_files, remove_empty_sub_dir


def parse_args(argv=None):
    hstr = '%prog custom help string'
    parser = OptionParser(hstr, description='custom description', version='%prog 1.0')
    parser.add_option('-s', '--source', action='store', dest='source', help='source package name')
    parser.add_option('-t', '--tools', action='store', dest='tools', default="",
                      help='tool package name list, join by `,`')
    parser.add_option('-d', '--dir', action='store', dest='dir', default=".", help='working dir')
    parser.add_option('-o', '--output', action='store', dest='output', default="pylib", help='target package name')
    parser.add_option('-c', '--compress', action='store', dest='compress', default=True,
                      help='compress the target package', )
    parser.add_option('-r', '--remove', action='store', dest='remove', default=False,
                      help='remove target dir if exists', )
    return parser.parse_args(argv if argv else sys.argv)


def compress_folder(source_folder, target_file_name):
    """

    :type target_file_name: str
    :type source_folder: str
    """
    logging.info("compress {} to {}".format(source_folder, target_file_name))
    compress_by_tar(source_folder, target_file_name, absolute_dir=False)


def copy_package(source_folder: str, target_folder: str, common_package_list: list,
                 current_dir: str):
    """
    if target_folder exists, it would remove target_folder first
    """

    def check_import_package(py_content: str, package_name_list: list) -> bool:
        _error_found = False

        for _package in package_name_list:
            str_list = ["from {} ".format(_package), "from {}.".format(_package)]

            for str_pattern in str_list:
                if py_content.find(str_pattern) > -1:
                    _error_found = True
                    logging.warning("found {} in py file".format(str_pattern))

        return _error_found

    def change_import_package(py_content: str, old_pkn: str, new_pkn: str):
        _package_found = False
        str_list = [
            "from {} ".format(old_pkn), "from {}.".format(old_pkn),
            "import {}.".format(old_pkn),
        ]

        for str_pattern in str_list:
            if py_content.find(str_pattern) > -1:
                _package_found = True
                py_content = py_content.replace(
                    str_pattern, str_pattern.replace(old_pkn, new_pkn, 1))

        str_pattern = "import {}\n".format(old_pkn)
        if py_content.find(str_pattern) > -1:
            _package_found = True
            py_content = py_content.replace(
                str_pattern, str_pattern.replace(old_pkn, new_pkn + " as {}".format(old_pkn), 1))

        return py_content, _package_found

    def change_common_import(_py_file: str, old_pkn: str, new_pkn: str) -> list:
        with open(_py_file, "r", encoding="utf-8") as f:
            py_content = f.read()

        common_package_using_list = []
        py_content, _ = change_import_package(py_content, old_pkn, new_pkn)
        for comm in common_package_list:
            py_content, _common_package_found = change_import_package(
                py_content, comm, "{}.{}".format(new_pkn, comm)
            )
            if _common_package_found:
                common_package_using_list.append(comm)

        package_list = [old_pkn, new_pkn]
        package_list.extend(common_package_list)

        other_package_names = [file_name for file_name in os.listdir(current_dir) if file_name not in set(package_list)]
        check_import_package(py_content, other_package_names)

        with open(_py_file, "w", encoding="utf-8") as fw:
            fw.write(py_content)

        return common_package_using_list

    def move_common_package(comm_pkn: str, new_pkn: str) -> list:
        # copy package
        copy_package_tree(
            os.path.join(current_dir, comm_pkn),
            os.path.join(target_folder, comm_pkn))

        _py_file_list = [file_name for file_name in
                         list_files(os.path.join(target_folder, comm_pkn)) if file_name.endswith(".py")]

        common_package_using_list = []
        for _py_file in _py_file_list:
            using_count = len(common_package_using_list)

            with open(_py_file, "r", encoding="utf-8") as f:
                _py_content = f.read()

            for comm in common_package_list:
                _py_content, _common_package_found = change_import_package(
                    _py_content, comm, "{}.{}".format(new_pkn, comm)
                )
                if _common_package_found:
                    common_package_using_list.append(comm)

            if len(common_package_using_list) > using_count:
                with open(_py_file, "w", encoding="utf-8") as fw:
                    fw.write(_py_content)

        return common_package_using_list

    def copy_package_tree(source: str, target: str):
        """
        only copy *.py in dir
        """
        shutil.copytree(source, target, symlinks=False)

        # just remove empty folder
        remove_empty_sub_dir(target, remove_input_folder=False)

    logging.info("copy package from {} to {}".format(source_folder, target_folder))
    old_package_name = os.path.basename(source_folder)
    new_package_name = os.path.basename(target_folder)

    # copy package dir
    if os.path.exists(target_folder):
        logging.warning("remove {}".format(target_folder))
        os.remove(target_folder)

    copy_package_tree(source_folder, target_folder)

    # change root import
    py_file_list = [file_name for file_name in list_files(target_folder) if file_name.endswith(".py")]
    use_comm_pkn = list()
    for py_file in py_file_list:
        use_comm_pkn.extend(change_common_import(py_file, old_package_name, new_package_name))

    # change common package
    use_comm_pkn = set(use_comm_pkn)
    if len(use_comm_pkn) == 0:
        logging.warning("not use {}".format(common_package_list))
        return

    # move common package
    to_move_comm_pkn = set(use_comm_pkn)
    moved_comm_pkn = set()
    while True:
        if not to_move_comm_pkn:
            break

        common_package = to_move_comm_pkn.pop()

        if common_package not in moved_comm_pkn:
            to_move_comm_pkn.union(
                set(move_common_package(common_package, new_package_name))
            )
            moved_comm_pkn.add(common_package)


def export_package(package_name: str, target_package_name: str, common_package_list: list,
                   is_gz=False, remove_target_dir=False, current_dir="."):
    """
    """
    # check args
    if package_name is None:
        raise ValueError("package_name cannot be null!")

    if target_package_name is None:
        raise ValueError("target_package_name cannot be null!")

    if is_gz is False and (target_package_name == package_name or target_package_name is None):
        raise ValueError("target_package_name cannot equal to package_name if is_gz=False!")

    if not os.path.exists(os.path.join(current_dir, package_name)):
        raise ValueError("sourc package [{}] not exists!".format(package_name))

    for comm_package in common_package_list:
        if not os.path.exists(os.path.join(current_dir, comm_package)):
            raise ValueError("common package [{}] not exists!".format(comm_package))

    if not os.path.exists(current_dir) or not os.path.isdir(current_dir):
        raise ValueError("current_dir[{}] is invalid!".format(current_dir))

    source_package_folder = os.path.join(current_dir, package_name)
    target_package_folder = os.path.join(current_dir, target_package_name)
    if not os.path.exists(source_package_folder):
        raise ValueError("source package not exists!")

    if target_package_folder != source_package_folder and os.path.exists(target_package_folder):
        if remove_target_dir:
            shutil.rmtree(target_package_folder)
        else:
            raise ValueError("target package folder exists!")

    # export folder to target folder
    copy_package(
        source_package_folder, target_package_folder,
        common_package_list=common_package_list,
        current_dir=current_dir,
    )

    # compress package
    if is_gz:
        compress_folder(target_package_folder, target_package_folder + ".tar.gz")


def export_package_helper():
    from pyxtools.basic_tools.log import global_init_logger
    global_init_logger()
    if sys.argv[0] == "-c":
        argv = list(sys.argv)
        argv[0] = __file__
        options, args = parse_args(argv)
    else:
        options, args = parse_args()
    export_package(package_name=options.source,
                   target_package_name=options.output,
                   common_package_list=[comm for comm in options.tools.split(",") if len(comm) > 0],
                   is_gz=options.compress,
                   remove_target_dir=options.remove,
                   current_dir=options.dir)


__all__ = ("export_package_helper",)



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import base64


def smart_decoder(raw_content, default_encoding_list=("utf-8", "gb18030")):
    """
    将字符串解码成unicode
    :type default_encoding_list: list of str
    :rtype: str
    :type raw_content: bytes
    """
    import chardet
    encoding = chardet.detect(raw_content).get("encoding", "utf-8")

    try:
        return raw_content.decode(encoding)
    except UnicodeDecodeError as e:
        for encoding in default_encoding_list:
            try:
                return raw_content.decode(encoding)
            except UnicodeDecodeError as e:
                pass
        raise e


def byte_to_string(byte_input: bytes, encoding="utf-8") -> str:
    """ encoding """
    return byte_input.decode(encoding)


def base64_to_string(base64_str):
    return base64.standard_b64decode(base64_str)


def string_to_base64(string):
    return base64.standard_b64encode(string)


__all__ = ("byte_to_string", "base64_to_string", "string_to_base64", "smart_decoder")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import time
import zipfile

import fnmatch
import functools
import os
import shutil
import tarfile


def list_files(folder):
    """

    :type folder: str
    :rtype: list of str
    """
    file_list = []
    for file_name in os.listdir(folder):
        full_file_name = os.path.join(folder, file_name)
        if os.path.isdir(full_file_name):
            file_list.extend(list_files(full_file_name))
        else:
            file_list.append(full_file_name)

    file_list.sort()
    return file_list


def remove_empty_sub_dir(folder, remove_input_folder=False):
    """

    :type remove_input_folder: bool
    :type folder: str
    :rtype: list of str
    """
    file_list = os.listdir(folder)

    for file_name in file_list:
        full_file_name = os.path.join(folder, file_name)
        if os.path.isdir(full_file_name):
            remove_empty_sub_dir(full_file_name, remove_input_folder=True)

    if not os.listdir(folder) and remove_input_folder:
        shutil.rmtree(folder)


def include_patterns(*patterns):
    """
    ref: https://stackoverflow.com/questions/42487578/python-shutil-copytree-use-ignore-function-to-keep-specific-files-types

    Factory function that can be used with copytree() ignore parameter.

    Arguments define a sequence of glob-style patterns
    that are used to specify what files to NOT ignore.
    Creates and returns a function that determines this for each directory
    in the file hierarchy rooted at the source directory when used with
    shutil.copytree().
    """

    def _ignore_patterns(path, names):
        keep = set(name for pattern in patterns
                   for name in fnmatch.filter(names, pattern))
        ignore = set(name for name in names
                     if name not in keep and not os.path.isdir(os.path.join(path, name)))
        return ignore

    return _ignore_patterns


def compress_by_tar(source, target_file_name, absolute_dir=True):
    """

    :type source: str
    :type target_file_name: str
    :type absolute_dir: bool
    """
    tar_filetype_list = {"tar", "gz"}
    file_type = os.path.basename(target_file_name).split(".")[-1]
    if file_type not in tar_filetype_list:
        raise ValueError("{} type not support!")

    def tar_add_folder(_folder, _tar, _relative_path=None):
        if _relative_path is None:
            _relative_path = _folder

        for _file_name in os.listdir(_folder):
            _full_file_name = os.path.join(_folder, _file_name)
            _current_rpath = os.path.join(_relative_path, _file_name)
            if os.path.isdir(_full_file_name):
                tar_add_folder(_full_file_name, _tar, _relative_path=_current_rpath)
            else:
                tar.add(_full_file_name, arcname=_current_rpath)

    with tarfile.open(target_file_name, "w:{}".format(file_type)) as tar:
        if os.path.isdir(source):
            if absolute_dir:
                tar_add_folder(source, tar, _relative_path=source)
            else:
                tar_add_folder(source, tar, _relative_path=os.path.basename(source))
        else:
            if absolute_dir:
                tar.add(source, arcname=source)
            else:
                tar.add(source, arcname=os.path.basename(source))


def check_file_exists(func):
    @functools.wraps(func)
    def check(*args, **kwargs):
        start_time = time.time()
        file_list = func(*args, **kwargs)
        result = [file_name for file_name in file_list if os.path.exists(file_name)]
        logging.info("check_file_exists cost time: {}s".format(time.time() - start_time))
        return result

    return check


def zip_compress_file(src_file_path, zip_file_path):
    with zipfile.ZipFile(zip_file_path, mode="w") as zf:
        zf.write(src_file_path, os.path.basename(src_file_path), compress_type=zipfile.ZIP_DEFLATED)


def get_base_name_of_file(path_or_file_name: str) -> str:
    if path_or_file_name.find("\\") > -1:
        path_or_file_name = path_or_file_name.replace("\\", "/")

    name_list = path_or_file_name.split("/")
    for name in name_list[::-1]:
        if name:
            return name

    return path_or_file_name


__all__ = ("get_base_name_of_file", "zip_compress_file", "check_file_exists", "list_files", "remove_empty_sub_dir",
           "compress_by_tar", "include_patterns")



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

# version info
VERSION = (2022, 5, 28, 1)
VERSION_STATUS = ""
VERSION_TEXT = ".".join(str(x) for x in VERSION) + VERSION_STATUS

setup(name="pyxtools",
      version=VERSION_TEXT,
      description="simple tool for Python3.6",
      long_description=open("README.rst", "r").read(),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Other Audience',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Scientific/Engineering',
          'Topic :: Software Development :: Libraries',
          'Topic :: Utilities',
      ],
      install_requires=[
          "requests",
          "chardet",
          "Pillow",
          "opencv-python",
          "numpy",
          "aiohttp",
          "matplotlib",
          "scipy",
      ],
      author="frkhit",
      url="https://github.com/frkhit/pyxtools",
      author_email="frkhit@gmail.com",
      license="MIT",
      packages=find_packages(),
      package_data={'': ["LICENSE", "README.rst", "MANIFEST.in"]},
      include_package_data=True, )



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import os
import shutil
import subprocess
import time
from decimal import getcontext, Decimal
from urllib.request import urlopen


def set_time_zone(tz: str = "Asia/Shanghai"):
    os.environ['TZ'] = tz
    time.tzset()


def set_proxy(http_proxy="http://127.0.0.1:8118"):
    """ 设置代理 """
    os.environ["https_proxy"] = http_proxy
    os.environ["HTTPS_PROXY"] = http_proxy
    os.environ["http_proxy"] = http_proxy
    os.environ["HTTP_PROXY"] = http_proxy


def clear_proxy_setting():
    """ 取消全局代理设置 """
    proxy_keys = {"HTTP_PROXY", "HTTPS_PROXY", "ALL_PROXY"}
    for key in proxy_keys:
        if key in os.environ:
            del os.environ[key]
        _key = key.lower()
        if _key in os.environ:
            del os.environ[_key]


def download_big_file(url, target_file_name):
    """
        使用python核心库下载大文件
        ref: https://stackoverflow.com/questions/1517616/stream-large-binary-files-with-urllib2-to-file
    """
    response = urlopen(url)
    chunk = 16 * 1024
    with open(target_file_name, 'wb') as f:
        while True:
            chunk = response.read(chunk)
            if not chunk:
                break
            f.write(chunk)


def download_big_file_with_wget(url, target_file_name):
    """
        使用wget下载大文件
        Note: 需要系统安装wget
    """

    download_process = subprocess.Popen(["wget", "-c", "-O", target_file_name, "'{}'".format(url)])

    download_process.wait()

    if not os.path.exists(target_file_name):
        raise Exception("fail to download file from {}".format(url))


def remove_path_or_file(path_or_file_name):
    """
        删除文件
    """
    if not os.path.exists(path_or_file_name):
        logging.warning("{} not exists!".format(path_or_file_name))
        return

    if os.path.isdir(path_or_file_name):
        # dir
        shutil.rmtree(path_or_file_name)
    else:
        # file
        os.remove(path_or_file_name)


def get_pretty_float(num: float, count: int = 2) -> str:
    """
        指定有效数字的科学计数法显示
    Args:
        num: float
        count: int
    """
    getcontext().prec = count
    return (Decimal(num) / Decimal(1)).to_eng_string()



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import argparse
import logging
import sys
from logging.handlers import TimedRotatingFileHandler, RotatingFileHandler

_inited_logger_args = None


def _clear_logger_handlers(_logger):
    while _logger.handlers:
        _handler = _logger.handlers[0]
        _logger.removeHandler(_handler)


def _logger_config_is_same(used_args, to_use_args) -> bool:
    if used_args[0] != to_use_args[0] or used_args[2] != to_use_args[2] or used_args[3] != to_use_args[3]:
        return False

    if isinstance(used_args[1], list) and isinstance(to_use_args[1], list):
        used_sorted_args = list(used_args[1])
        used_sorted_args.sort()
        to_use_sorted_args = list(to_use_args[1])
        to_use_sorted_args.sort()
        return "<<>>".join(to_use_sorted_args) == "<<>>".join(used_sorted_args)

    return used_args[1] == to_use_args[1]


def get_time_rotating_handler(log_file: str) -> logging.FileHandler:
    fh = TimedRotatingFileHandler(log_file, when='midnight', interval=1, backupCount=30)
    return fh


def get_rotating_file_handler(log_file: str) -> logging.FileHandler:
    fh = RotatingFileHandler(log_file, maxBytes=2 * 1024 * 1024 * 1024, backupCount=30)
    return fh


def global_init_logger(logger_level=logging.INFO, log_file: str = None, reset_logger_name_list: list = None,
                       log_file_handler: logging.FileHandler = None):
    global _inited_logger_args

    if _inited_logger_args is not None and _logger_config_is_same(_inited_logger_args, (
            log_file, reset_logger_name_list, logger_level, log_file_handler)):
        logging.warning("init logger run before!")
        return

    # formatter
    _default_formatter = logging.Formatter("%(asctime)s %(filename)s - %(name)s - %(levelname)s - %(message)s", None)

    # reset other logger
    if reset_logger_name_list is not None:
        for logger_name in reset_logger_name_list:
            _clear_logger_handlers(logging.getLogger(logger_name))

    # reset root logger
    root_logger = logging.getLogger()
    _clear_logger_handlers(root_logger)

    # console handler
    _console_handler = logging.StreamHandler(sys.stdout)
    _console_handler.setFormatter(_default_formatter)
    root_logger.addHandler(_console_handler)

    # file handler
    if log_file_handler is None and log_file is not None:
        log_file_handler = logging.FileHandler(log_file)

    if log_file_handler is not None:
        log_file_handler.setFormatter(_default_formatter)
        root_logger.addHandler(log_file_handler)

    root_logger.setLevel(logger_level)

    if reset_logger_name_list is None:
        _inited_logger = (log_file, None, logger_level)
    else:
        _inited_logger = (log_file, list(reset_logger_name_list), logger_level)


def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Unsupported value encountered.')


__all__ = ("global_init_logger", "get_time_rotating_handler", "get_rotating_file_handler", "str2bool")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import logging
import os
import shutil

CRLF_VALUE = {"lf": "\n", "crlf": "\r\n", "cr": "\r"}


def change_crlf_mode_of_file(file_name, dst_lf):
    tmp_file = file_name + ".bak"
    shutil.copy(file_name, tmp_file, follow_symlinks=False)

    with open(tmp_file, 'r', newline=None, encoding="utf-8") as infile, \
            open(file_name, 'w', newline=dst_lf, encoding="utf-8") as outfile:
        outfile.writelines(infile.readlines())

    os.remove(tmp_file)


def change_to_lf(file_list: list):
    for file_name in file_list:
        try:
            change_crlf_mode_of_file(file_name, "\n")
        except Exception as e:
            logging.error(e, exc_info=True)


def change_to_crlf(file_list: list):
    for file_name in file_list:
        try:
            change_crlf_mode_of_file(file_name, "\r\n")
        except Exception as e:
            logging.error(e, exc_info=True)


__all__ = ("change_to_crlf", "change_to_lf", "change_crlf_mode_of_file", "CRLF_VALUE")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import random


def iter_list_with_size(src_list: list, size: int):
    """
        src_list would be modified when running
    """
    n_part = len(src_list) // size + 1
    while n_part >= 0:
        n_part -= 1
        part_src_list = src_list[:size]
        if part_src_list:
            yield part_src_list
            del src_list[:size]
        else:
            break


def random_choice(src_list, k: int = 1, unique: bool = True):
    if unique is False:
        return random.choices(src_list, k=k)

    assert k <= len(src_list)
    tmp_list = list(src_list)
    random.seed()  # todo: random.seed(0) is wrong here!
    random.shuffle(tmp_list)
    random.shuffle(tmp_list)
    return tmp_list[:k]


__all__ = ("iter_list_with_size", "random_choice")



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import os
import unittest

import numpy as np

from pyxtools.tmp_tools.sqlalchemy_demo import FeatureRecord, get_sqlite_engine, get_sqlite_session


class TestSqlAlchemy(unittest.TestCase):
    def setUp(self):
        self.file = "./tmp.dat"
        FeatureRecord.sqlite_file = self.file
        FeatureRecord.init_table_object()

    def tearDown(self):
        if os.path.exists(self.file):
            os.remove(self.file)

    def testClose(self):
        engine = get_sqlite_engine(sqlite_file=self.file)
        db_session = get_sqlite_session(engine=engine)
        self.assertEqual(id(db_session.bind), id(engine))

    def testInfo(self):
        db_session = FeatureRecord.get_session()
        train, test = FeatureRecord.get_info(db_session)
        self.assertTrue(isinstance(train, int))
        self.assertTrue(isinstance(test, int))
        db_session.close()

    def testAdd(self):
        db_session = FeatureRecord.get_session()

        train, test = FeatureRecord.get_info(db_session)
        count = train + test

        # train data
        for index, label in enumerate([1, -1]):
            feature = np.random.random((1, 1024))
            instance = FeatureRecord.add_feature(db_session=db_session, image_id="a.jpg", label=label,
                                                 feature=feature)

            train, test = FeatureRecord.get_info(db_session)
            print("train is {}, test is {}, count is {}".format(train, test, count))
            # self.assertTrue((count + 1) == (train + test))
            count = train + test

            result_list = FeatureRecord.list_feature(db_session=db_session, uid_list=[instance.id])
            self.assertTrue(len(result_list) == 1)
            (obj_id, obj_image_id, obj_label, obj_feature) = result_list[0]
            self.assertTrue(obj_id == instance.id)
            self.assertTrue(obj_image_id == instance.image_id)
            self.assertTrue(obj_label == instance.label)
            self.assertTrue(type(obj_feature) == type(feature))
            self.assertTrue(np.sum(obj_feature) == np.sum(feature))

        db_session.close()



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import cProfile
import logging
import pstats
import time

import functools
import re


class TimeCostHelper(object):
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self._time_start = time.time()
        self._last_time = time.time()

    def dot(self, info_format: str = None):
        now = time.time()
        time_cost_seconds = now - self._last_time
        self._last_time = now
        if info_format:
            self.logger.info(info_format.format(time_cost_seconds))
        else:
            self.logger.info("Time cost {}s".format(time_cost_seconds))

    def sum(self, info_format: str = None):
        now = time.time()
        time_cost_seconds = now - self._time_start
        if info_format:
            self.logger.info(info_format.format(time_cost_seconds))
        else:
            self.logger.info("Total time cost {}s".format(time_cost_seconds))


def time_cost(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        logging.info("<{}> cost time: {}s".format(func.__name__, time.time() - start_time))
        return result

    return wrapper


def c_profile_demo():
    print(re)  # re must import
    cProfile.run('re.compile("foo|bar")')


def do_c_profile(prof_file: str):
    """
    ref: https://zhuanlan.zhihu.com/p/24495603
    Decorator for function profiling.
    """

    def wrapper(func):
        def profiled_func(*args, **kwargs):
            # Flag for do profiling or not.
            profile = cProfile.Profile()
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()
            # Sort stat by internal time.
            sort_by = "tottime"
            ps = pstats.Stats(profile).sort_stats(sort_by)
            ps.dump_stats(prof_file)
            return result

        return profiled_func

    return wrapper


def parse_c_profile_file(prof_file: str):
    p = pstats.Stats(prof_file)
    p.strip_dirs().sort_stats("cumtime").print_stats(10, 1.0, ".*")


__all__ = ("time_cost", "do_c_profile", "parse_c_profile_file", "TimeCostHelper")
